import flet as ft

def error_banner(gui,message):
    gui.page.open(ft.SnackBar(
        ft.Text(message)))
    gui.page.update()

def handle_image_switch_mask_on(gui):
    """
    loads the mask into the GUI if the switch is on
    Args:
        gui (GUI):the current GUI object

    """
    if gui.switch_mask.value:
        image = gui.csp.image_id
        bfc=gui.csp.config.get_bf_channel()

       #case: mask was created during segmentation
        if image in gui.csp.mask_paths and bfc in gui.csp.mask_paths[image]:
            #if the image to the bright-field channel was not generated before
            if image not in gui.mask.mask_outputs or bfc not in gui.mask.mask_outputs[image]:
                gui.mask.load_mask_into_canvas()

            #loads mask into container
            insert_mask(gui,image,bfc)
        else:
            #case off
            error_banner(gui,f"There is no mask for {gui.csp.image_id} with bright-field channel {bfc} generated ")
            gui.canvas.container_mask.visible = False
            gui.switch_mask.value=False
    else:
        gui.canvas.container_mask.visible = False

    gui.page.update()

def handle_mask_update(gui):
    """
    The method is executed if the mask was updated in the drawing window.
    Loads the updated mask in the flet canvas.

    Args:
        gui (GUI): the current GUI object
    """

    image= gui.csp.image_id
    bfc=gui.csp.config.get_bf_channel()

    if gui.switch_mask.value:
        gui.mask.load_mask_into_canvas()
        insert_mask(gui,image,bfc)
    else:
        return

def reset_mask(gui,image_id,bf_channel):
    if image_id in gui.mask.mask_outputs and bf_channel in gui.mask.mask_outputs[image_id]:
        del gui.mask.mask_outputs[image_id][bf_channel]


def insert_mask(gui, image,bfc):
    mask = gui.mask.mask_outputs[image][bfc]
    gui.canvas.container_mask.content.src_base64= mask
    gui.canvas.container_mask.update()
    gui.canvas.container_mask.visible = True



import asyncio
import threading
import numpy as np
from PyQt5.QtCore import Qt, QPoint, pyqtSignal, QObject, QTimer, QPointF
from PyQt5.QtGui import QImage, QPixmap, QPen, QColor
from PyQt5.QtWidgets import QApplication, QLabel, QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QGraphicsScene, \
    QGraphicsView, QMainWindow, QGraphicsLineItem, QCheckBox
import sys

from cellsepi.backend.drawing_window.drawing_util import mask_shifting, bresenham_line, search_free_id, fill_polygon_from_outline, \
    find_border_pixels, trace_contour


class MyQtWindow(QMainWindow):
    """
    Main PyQt window for drawing left_tools and deleting cells.

    Attributes:
        canvas: DrawingCanvas object for displaying and interacting with the mask.
        canvas_dummy: says if the canvas is a dummy or not.
        tools_widget: Container for left_tools on the right side.
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Mask Editing")
        self.check_shifting = QCheckBox("Cell ID shifting")
        self.check_shifting.setStyleSheet(
            "font-size: 16px; color:#000000; padding: 10px 20px; margin-bottom: 10px; background-color: #F5F5F5; border: 1px solid #CCCCCC; border-radius: 5px;")
        self.canvas_dummy = True

        self.canvas = QWidget()

        #Main layout with canvas and left_tools
        central_widget = QWidget()
        self.main_layout = QHBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)  #Remove margins on all sides

        self.main_layout.addWidget(self.canvas, stretch=3)

        #Add left_tools box to the right
        self.tools_widget = QWidget()
        tools_layout = QVBoxLayout()
        tools_layout.setContentsMargins(10, 10, 10, 10)  #Consistent padding
        tools_layout.setAlignment(Qt.AlignTop)  #Align left_tools to the top

        #Add title to the left_tools box
        title = QLabel("Tools")
        title.setStyleSheet(
            "font-size: 20px; font-weight: bold; color: #333; padding: 10px; text-align: center; background-color: #EDEDED; border-radius: 5px;")
        tools_layout.addWidget(title)



        #Add buttons to the left_tools box
        self.mask_toggle_button = QPushButton("Mask: ON")
        self.mask_toggle_button.setCheckable(True)
        self.mask_toggle_button.setChecked(True)
        self.mask_toggle_button.setStyleSheet(
            "font-size: 16px; color: #000000; padding: 10px 20px; margin-bottom: 10px; background-color: #F5F5F5; border: 1px solid #CCCCCC; border-radius: 5px;")
        self.mask_toggle_button.clicked.connect(self.toggle_mask)
        tools_layout.addWidget(self.mask_toggle_button)

        self.draw_toggle_button = QPushButton("Drawing: ON")
        self.draw_toggle_button.setCheckable(True)
        self.draw_toggle_button.setChecked(True)
        self.draw_toggle_button.setStyleSheet(
            "font-size: 16px; color:#000000; padding: 10px 20px; margin-bottom: 10px; background-color: #F5F5F5; border: 1px solid #CCCCCC; border-radius: 5px;")
        self.draw_toggle_button.clicked.connect(self.toggle_draw_mode)
        tools_layout.addWidget(self.draw_toggle_button)

        self.delete_toggle_button = QPushButton("Delete Mode: OFF")
        self.delete_toggle_button.setCheckable(True)
        self.delete_toggle_button.setStyleSheet(
            "font-size: 16px; color: #000000; padding: 10px 20px; margin-bottom: 10px; background-color: #F5F5F5; border: 1px solid #CCCCCC; border-radius: 5px;")
        self.delete_toggle_button.clicked.connect(self.toggle_delete_mode)
        tools_layout.addWidget(self.delete_toggle_button)


        tools_layout.addWidget(self.check_shifting, alignment=Qt.AlignCenter)

        self.restore_button = QPushButton("Undo")
        self.restore_button.clicked.connect(self.restore_cell)
        self.restore_button.setStyleSheet("""
            QPushButton:disabled {
                background-color: lightgray;
                color: gray;
                border: 1px solid darkgray;
            }
            QPushButton:enabled {
                border: 1px solid darkgray;
                color: #000000
            }
        """)
        self.restore_button.setEnabled(False)
        tools_layout.addWidget(self.restore_button)

        self.redo_button = QPushButton("Redo")
        self.redo_button.clicked.connect(self.redo_delete)
        self.redo_button.setStyleSheet("""
            QPushButton:disabled {
                background-color: lightgray;
                color: gray;
                border: 1px solid darkgray;
            }
            QPushButton:enabled {
                border: 1px solid darkgray;
                color: #000000
            }
        """)
        self.redo_button.setEnabled(False)
        tools_layout.addWidget(self.redo_button)

        self.tools_widget.setLayout(tools_layout)
        self.tools_widget.setStyleSheet(
            "background-color: #FAFAFA; border-left: 2px solid #E0E0E0;")  #Subtle border and clean background
        self.tools_widget.setFixedWidth(250)
        self.main_layout.addWidget(self.tools_widget)

        central_widget.setLayout(self.main_layout)
        self.setCentralWidget(central_widget)

        #screen layout
        screen = QApplication.primaryScreen()
        screen_geometry = screen.geometry()
        window_geometry = self.geometry()
        x = (screen_geometry.width() - window_geometry.width()) // 2
        y = (screen_geometry.height() - window_geometry.height()) // 2

        self.move(x, y)

    def restore_cell(self):
        if not self.canvas_dummy:
            self.canvas.restore_cell()

    def redo_delete(self):
        if not self.canvas_dummy:
            self.canvas.redo_delete()

    def set_queue_image(self, mask_color, outline_color, opacity, bf_channel, mask_paths, image_id, adjusted_image_path, conn,
                        mask_path, channel_id, channel_prefix,slice_id,max_slice_id):
        """
        Sets the current selected mask and image into the MyQtWindow, replacing the canvas with the current parameters.
        Also updates the window title to include the image_id.

        Arguments:
            mask_color: The color to use for the mask
            outline_color: The color to use for the outline
            opacity: The opacity of the mask
            bf_channel: The current bf channel
            mask_paths: The paths to each masks
            image_id: The id of the current selected image
            adjusted_image_path: The path to the adjusted image, which is the current image with the current adjustments like brightness and contrast adjustments
            conn: The pipeline connection to the Flet window to communicate with the main window
            mask_path: The path to the current selected mask
            channel_id: The id of the current selected channel
            channel_prefix: The prefix of the channel
            slice_id: The id of the current selected slice for 3d images
            max_slice_id: The max slice id for the 3d image
        """
        #Update the window title with the current image's ID and channel ID.
        self.setWindowTitle(f"Mask Editing - {image_id}{channel_prefix}{channel_id}")

        #if canvas is only a dummy create a new_canvas
        if self.canvas_dummy:
            new_canvas = DrawingCanvas(mask_color, outline_color, opacity, bf_channel, mask_paths, image_id, adjusted_image_path,
                                       self.check_shifting, conn, mask_path,slice_id,max_slice_id, True, False,mask_show=self.mask_toggle_button.isChecked())
        else:
            #check if the current image_id and bf_channel equal to the values in canvas than only update the background image (adjusted_image)
            if image_id == self.canvas.image_id and bf_channel == self.canvas.bf_channel and (slice_id==self.canvas.slice_id):
                if opacity == self.canvas.opacity:
                    self.canvas.adjusted_image_path = adjusted_image_path
                    self.canvas.load_image_to_scene()
                else:
                    self.canvas.opacity = opacity
                    self.canvas.adjusted_image_path = adjusted_image_path
                    self.canvas.load_image_to_scene()
                    self.canvas.load_mask_to_scene()
                return  #return because no new canvas, so no need for replacing the old canvas with the new and update the connections
            else:
                #new mask and image so need for new_canvas
                new_canvas = DrawingCanvas(mask_color, outline_color, opacity, bf_channel, mask_paths, image_id,
                                           adjusted_image_path,
                                           self.check_shifting, conn, mask_path,slice_id,max_slice_id, self.canvas.draw_mode,
                                           self.canvas.delete_mode,mask_show=self.mask_toggle_button.isChecked())

        #replace the old canvas with the new one and update the window
        self.restore_button.setEnabled(False)
        self.redo_button.setEnabled(False)
        self.canvas_dummy = False
        self.main_layout.replaceWidget(self.canvas, new_canvas)
        self.canvas.deleteLater()
        self.canvas = new_canvas
        self.canvas.update()
        self.main_layout.update()
        #Connect signals to update the state of the restore and redo buttons.
        self.canvas.restoreAvailabilityChanged.connect(lambda available: self.restore_button.setEnabled(available))
        self.canvas.redoAvailabilityChanged.connect(lambda available: self.redo_button.setEnabled(available))

        QTimer.singleShot(0, lambda: self.canvas.fitInView(self.canvas.sceneRect(), Qt.KeepAspectRatio)) #sets the right aspect ratio to the canvas

    def toggle_draw_mode(self):
        """
        Drawing Button functionality
        """
        if not self.canvas_dummy:
            if self.draw_toggle_button.isChecked():
                self.draw_toggle_button.setText("Drawing : ON")
                self.delete_toggle_button.setChecked(False)
                self.delete_toggle_button.setText("Delete Mode: OFF")
                self.canvas.set_delete_mode(False)
            else:
                self.draw_toggle_button.setText("Drawing : OFF")
            self.canvas.toggle_draw_mode()

    def toggle_delete_mode(self):
        """
        Toggle delete mode when the button is clicked.
        """
        if not self.canvas_dummy:
            if self.delete_toggle_button.isChecked():
                self.delete_toggle_button.setText("Delete Mode: ON")
                self.draw_toggle_button.setChecked(False)
                self.draw_toggle_button.setText("Drawing : OFF")
                self.canvas.set_draw_mode(False)
            else:
                self.delete_toggle_button.setText("Delete Mode: OFF")
            self.canvas.toggle_delete_mode()

    def toggle_mask(self):
        """
        Toggles if the mask is shown or not.
        """
        if not self.canvas_dummy:
            if self.mask_toggle_button.isChecked():
                self.mask_toggle_button.setText("Mask : ON")
                self.canvas.mask_item.setVisible(True)
                self.canvas.mask_show = True
            else:
                self.mask_toggle_button.setText("Mask : OFF")
                self.canvas.mask_item.setVisible(False)
                self.canvas.mask_show = False

    def resizeEvent(self, event):
        self.canvas.fitInView(self.canvas.sceneRect(), Qt.KeepAspectRatio)


class Updater(QObject):
    """
    Handles the signals that come from the queue.
    """
    update_signal = pyqtSignal(object, object)  #Signal for new main_image
    close_signal = pyqtSignal(object, object)  #Signal to close the drawing window
    delete_signal = pyqtSignal(object, )  #Signal that the main_image mask got deleted
    refresh_signal = pyqtSignal()  #Signal that the main_image mask got deleted
    color_change_signal = pyqtSignal(object, object) #Signal that the colors got changed
    hide_signal = pyqtSignal()

    def __init__(self, window):
        super().__init__()
        self.update_signal.connect(self.handle_update)
        self.close_signal.connect(self.handle_close)
        self.delete_signal.connect(self.handle_delete)
        self.refresh_signal.connect(self.handle_refresh)
        self.color_change_signal.connect(self.update_color)
        self.window: MyQtWindow = window
        self.hide_signal.connect(self.handle_hide)

    def handle_update(self, data, conn):
        """
        If the update signal is received, update the window accordingly.
            """
        mask_color, outline_color, opacity, bf_channel, mask_paths, image_id, adjusted_image_path, mask_path, channel_id, channel_prefix = data[:10]

        slice_id = data[10] if len(data) > 10 else None #optional 3D slice_id
        max_slice_id = data[11] if len(data) > 11 else None #optional 3d max_slice_id
        self.window.set_queue_image(mask_color, outline_color, opacity, bf_channel, mask_paths, image_id, adjusted_image_path,
                                    conn, mask_path, channel_id, channel_prefix,slice_id,max_slice_id)
        self.window.setVisible(True)
        self.window.raise_()
        self.window.activateWindow()

    def handle_close(self, app, running):
        """
        If the close signal is received, close the process.
        """
        self.window.hide()
        running[0] = False
        app.quit()

    def handle_hide(self):
        self.window.hide()

    def handle_delete(self, app):
        """
        If the delete signal is received, closes the process, but lets it restart invisible.
        """
        self.window.hide()
        app.quit()

    def handle_refresh(self):
        """
        If the refresh signal is received, refreshes the mask in canvas from disc.
        """
        if not self.window.canvas_dummy:
            self.window.canvas.store_mask_and_update()

    def update_color(self, mask_color, outline_color):
        """
        If the color change signal is received, update the color of mask and outline.
        """
        if not self.window.canvas_dummy:
            self.window.canvas.mask_color = mask_color
            self.window.canvas.outline_color = outline_color
            self.window.canvas.load_mask_to_scene()

def open_qt_window(queue, conn):
    """
    Opens the Qt window invisibly and updates it based on signals received through the queue (e.g., close, update, color change, refresh).

    Arguments:
        queue: The connection to the main application, which sends information to the Qt window.
            - mask_color, outline_color, bf_channel, mask_paths, image_id, adjusted_image_path, mask_path, channel_id, channel_prefix: Makes the window visible or updates the canvas image if already visible.
            - "close": Closes the Qt window.
            - "delete_mask": Closes the Qt window, because the mask got deleted.
            - "refresh_mask": Refreshes the current mask from disk.
            - "color_change", mask_color, outline_color: Changes the mask and outline color.
        conn: A communication pipeline to the main application for sending information to the main window.
            - "update_mask": Updates the current mask from disk in the main window.
            - "close": Closes the pipeline listener on the main window side.
    """
    app = QApplication(sys.argv)
    running = [True]
    thread = None
    while running[0]:
        window = MyQtWindow()
        window.setVisible(False)
        updater = Updater(window)
        conn.send("ready")

        def background_listener():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def queue_listener():
                while running[0]:
                    data = await asyncio.to_thread(queue.get)
                    if data == "close":
                        updater.close_signal.emit(app, running)
                        break
                    elif data == "delete_mask":
                        updater.delete_signal.emit(app)
                    elif data == "refresh_mask":
                        updater.refresh_signal.emit()
                    elif data == "hide":
                        updater.hide_signal.emit()
                    elif data == "close_thread":
                        break
                    elif isinstance(data, (tuple, list)) and data[0] == "color_change":
                        _, mask_color, outline_color = data
                        updater.color_change_signal.emit(mask_color, outline_color)
                    else:
                        updater.update_signal.emit(data, conn)

            try:
                loop.run_until_complete(queue_listener())
            finally:
                loop.stop()
                loop.close()

        thread = threading.Thread(target=background_listener, daemon=True)
        thread.start()
        app.exec_()
        if running[0]:
            queue.put("close_thread")
        if thread is not None and thread.is_alive():
            thread.join()
        window.close()
        window.deleteLater()
    conn.send("close")
    sys.exit(0)


class DrawingCanvas(QGraphicsView):
    """
    Class for displaying and interacting with images (background, mask).
    Includes delete functionality for cells and supports undo (restore) and redo of cell deletion.
    """
    #Signals to indicate whether a restore or redo action is available.
    restoreAvailabilityChanged = pyqtSignal(bool)
    redoAvailabilityChanged = pyqtSignal(bool)

    def __init__(self, mask_color, outline_color,opacity, bf_channel, mask_paths, image_id, adjusted_image_path, check_box,
                 conn, mask_path,slice_id,max_slice_id, draw_mode=True, delete_mode=False,mask_show=True):
        super().__init__()

        self.mask_color = mask_color
        self.outline_color = outline_color
        self.opacity = opacity
        self.bf_channel = bf_channel
        self.mask_paths = mask_paths
        self.image_id = image_id
        self.slice_id= slice_id
        self.max_slice_id = max_slice_id
        self.adjusted_image_path = adjusted_image_path
        self.mask_path = mask_path
        self.scene = QGraphicsScene(self)
        self.setScene(self.scene)
        self.draw_mode = draw_mode
        self.last_point = QPoint()
        self.start_point = None
        self.drawing = False
        self.delete_mode = delete_mode
        self.image_array = None
        self.mask_item = None
        self.background_item = None
        self.mask_data = None
        self.cell_history = []  #Each entry: (old_state, new_state)
        self.redo_history = []  #Each entry: (old_state, new_state)
        self.check_box = check_box
        self.points = []  #saves all points in the drawing
        self.conn = conn
        self.mask_show = mask_show
        self.load_mask_to_scene()
        self.load_image_to_scene()

    def toggle_draw_mode(self):
        self.draw_mode = not self.draw_mode

    def toggle_delete_mode(self):
        self.delete_mode = not self.delete_mode

    def set_draw_mode(self, value: bool):
        self.draw_mode = value

    def set_delete_mode(self, value: bool):
        self.delete_mode = value

    def is_point_within_image(self, point):
        """
        Check if a point is within the boundaries of the image.
        """
        if self.image_array is None:
            return False  #No image loaded
        x, y = int(point.x()), int(point.y())
        return 0 <= x < self.image_array.shape[1] and 0 <= y < self.image_array.shape[0]

    def mousePressEvent(self, event):
        """
        Handle mouse click events for drawing or deleting cells.
        """
        if self.draw_mode:
            if event.button() == Qt.LeftButton:
                current_point = self.mapToScene(event.pos())
                if not self.is_point_within_image(current_point):
                    x,y = self.clamp_to_image_bounds(current_point)
                    current_point = QPointF(x, y)

                self.drawing = True
                self.last_point = current_point

                if self.start_point is None:
                    self.start_point = current_point
        elif self.delete_mode:
            pos = event.pos()
            scene_pos = self.mapToScene(pos)
            cell_id = self.get_cell_id_from_position(scene_pos)
            if cell_id:
                self.delete_cell(cell_id)
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """
        What happens, if you move the clicked mouse
        """
        if self.draw_mode:
            current_point = self.mapToScene(event.pos())

            if self.is_point_within_image(current_point):
                x, y = int(current_point.x()), int(current_point.y())
            else:
                x, y = self.clamp_to_image_bounds(current_point)

            if self.last_point:
                line_item = QGraphicsLineItem(self.last_point.x(), self.last_point.y(), x, y)
                r, g, b = self.outline_color
                pen = QPen(QColor(r, g, b), 2, Qt.SolidLine)
                line_item.setPen(pen)
                self.scene.addItem(line_item)

            self.last_point = QPointF(x, y)

    def clamp_to_image_bounds(self, point):
        """
        If mouse goes out of image, the line stays in the image bounds.
        """
        x, y = int(point.x()), int(point.y())
        x = max(0, min(x, self.image_array.shape[1] - 1))
        y = max(0, min(y, self.image_array.shape[0] - 1))

        return x, y

    def mouseReleaseEvent(self, event):
        """
        If you release the left click.
        """
        if self.draw_mode and self.drawing:
            self.drawing = False

            # connect last and first point in pic
            if self.start_point and self.last_point:
                first_inside = self.clamp_to_image_bounds(self.start_point)
                last_inside = self.clamp_to_image_bounds(self.last_point)

                line_item = QGraphicsLineItem(first_inside[0], first_inside[1],
                                              last_inside[0], last_inside[1])
                r, g, b = self.outline_color
                pen = QPen(QColor(r, g, b), 2, Qt.SolidLine)
                line_item.setPen(pen)
                self.scene.addItem(line_item)
            self.start_point = None
            self.last_point = None

            self.update()
            self.add_drawn_cell()
        else:
            super().mouseReleaseEvent(event)

    def get_cell_id_from_position(self, position):
        """
        Get the cell ID from the clicked position.
        """
        x, y = int(position.x()), int(position.y())
        if 0 <= x < self.image_array.shape[1] and 0 <= y < self.image_array.shape[0]:
            return self.image_array[y, x]
        return None

    def store_mask_and_update(self):
        mask = self.mask_data["masks"]
        outline = self.mask_data["outlines"]

        #Save current state of the cell for restoration (undo)
        mask_old = mask.copy()
        outline_old = outline.copy()
        self.load_mask_to_scene()
        mask = self.mask_data["masks"]
        outline = self.mask_data["outlines"]

        if mask.ndim == 3:
            mask = np.take(mask, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        if outline.ndim == 3:
            outline = np.take(outline, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        # Save current state of the cell for restoration (undo)
        new_mask = mask.copy()
        new_outline = outline.copy()
        self.cell_history.append(((mask_old, outline_old), (new_mask, new_outline)))

        self.restoreAvailabilityChanged.emit(len(self.cell_history) > 0)
        self.redoAvailabilityChanged.emit(len(self.redo_history) > 0)

    def delete_cell(self, cell_id):
        """
        Delete the specified cell by updating the mask data.
        Also does not clear stored redo history, enabling multiple redo levels.
        """
        mask_path = self.mask_paths[self.image_id][self.bf_channel]
        self.mask_data = np.load(mask_path, allow_pickle=True).item()

        mask = self.mask_data["masks"]
        outline = self.mask_data["outlines"]

        if mask.ndim == 3:
            mask = np.take(mask, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        if outline.ndim == 3:
            outline = np.take(outline, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        #Save current complete state before deletion
        old_state = (mask.copy(), outline.copy())

        #Update the mask and outline (delete the cell)
        cell_mask = (mask == cell_id).copy()
        cell_outline = (outline == cell_id).copy()
        mask[cell_mask] = 0
        outline[cell_outline] = 0
        if self.check_box.isChecked():
            mask_shifting(self.mask_data, cell_id,self.slice_id)

        #Save new state after deletion
        mask_3d = None
        outline_3d = None
        if self.slice_id is not None:
            mask_3d = self.mask_data["masks"]
            outline_3d = self.mask_data["outlines"]

            if mask_3d.ndim == 3:
                mask_3d[int(self.slice_id),:,:] = mask

            if outline_3d.ndim == 3:
                outline_3d[int(self.slice_id),:,:] = outline

        #Redo: restore the new state
        np.save(mask_path, {"masks": mask if self.slice_id is None else mask_3d, "outlines": outline if self.slice_id is None else outline_3d}, allow_pickle=True)
        new_state = (mask.copy(), outline.copy())

        self.load_mask_to_scene()
        self.conn.send("update_mask")

        #Save the state transition in undo history
        self.cell_history.append((old_state, new_state))
        self.restoreAvailabilityChanged.emit(len(self.cell_history) > 0)
        self.redoAvailabilityChanged.emit(len(self.redo_history) > 0)

    def restore_cell(self):
        """
        Restore the most recent state (undo the deletion or drawing).
        Also stores the undone action in a redo history so that the change can be re-applied.
        """
        if not self.cell_history:
            #No state to restore; update button state via signal and do nothing.
            self.restoreAvailabilityChanged.emit(False)
            return

        mask_path = self.mask_paths[self.image_id][self.bf_channel]
        #Retrieve the last state transition from undo history
        old_state, new_state = self.cell_history.pop()

        mask_3d = None
        outline_3d = None
        if self.slice_id is not None:
            mask_data = np.load(mask_path, allow_pickle=True).item()
            mask_3d = mask_data["masks"]
            outline_3d = mask_data["outlines"]

            if mask_3d.ndim == 3:
                mask_3d[ int(self.slice_id),:,:] = old_state[0]

            if outline_3d.ndim == 3:
                outline_3d[int(self.slice_id),:,:] = old_state[1]

        #Redo: restore the new state
        np.save(mask_path, {"masks": old_state[0] if self.slice_id is None else mask_3d, "outlines": old_state[1] if self.slice_id is None else outline_3d}, allow_pickle=True)
        self.load_mask_to_scene()
        self.conn.send("update_mask")

        #Save the undone action in redo history
        self.redo_history.append((old_state, new_state))
        self.restoreAvailabilityChanged.emit(len(self.cell_history) > 0)
        self.redoAvailabilityChanged.emit(len(self.redo_history) > 0)

    def redo_delete(self):
        """
        Re-apply the deletion (redo the undone deletion or drawing).
        """
        if not self.redo_history:
            self.redoAvailabilityChanged.emit(False)
            return
        mask_path = self.mask_paths[self.image_id][self.bf_channel]
        #Retrieve the state transition from redo history
        old_state, new_state = self.redo_history.pop()

        mask_3d = None
        outline_3d = None
        if self.slice_id is not None:
            mask_data = np.load(mask_path, allow_pickle=True).item()
            mask_3d = mask_data["masks"]
            outline_3d = mask_data["outlines"]

            if mask_3d.ndim == 3:
                mask_3d[int(self.slice_id),:, :] = new_state[0]

            if outline_3d.ndim == 3:
                outline_3d[int(self.slice_id),:, :] = new_state[1]

        #Redo: restore the new state
        np.save(mask_path, {"masks": new_state[0] if self.slice_id is None else mask_3d, "outlines": new_state[1] if self.slice_id is None else outline_3d}, allow_pickle=True)
        self.load_mask_to_scene()
        self.conn.send("update_mask")

        #After redoing, push it back to undo history
        self.cell_history.append((old_state, new_state))
        self.restoreAvailabilityChanged.emit(len(self.cell_history) > 0)
        self.redoAvailabilityChanged.emit(len(self.redo_history) > 0)

    def load_mask_to_scene(self):
        """
        Load the mask and display it on the scene.
        """
        if self.image_id not in self.mask_paths:
            self.mask_paths[self.image_id] = {}

        if self.bf_channel not in self.mask_paths[self.image_id]:
            pixmap = QPixmap(self.adjusted_image_path)
            if self.slice_id is None:
                empty_mask = {
                    "masks": np.zeros((pixmap.width(), pixmap.height()), dtype=np.uint8),
                    "outlines": np.zeros((pixmap.width(), pixmap.height()), dtype=np.uint8)
                }
            else:
                empty_mask = {
                    "masks": np.zeros((self.max_slice_id,pixmap.width(), pixmap.height()), dtype=np.uint8),
                    "outlines": np.zeros((self.max_slice_id,pixmap.width(), pixmap.height()), dtype=np.uint8)
                }

            np.save(self.mask_path, empty_mask)
            self.conn.send(f"new_mask.{self.image_id}")
            self.mask_paths[self.image_id][self.bf_channel] = self.mask_path

        mask_path = self.mask_paths[self.image_id][self.bf_channel]

        self.mask_data = np.load(mask_path, allow_pickle=True).item()
        mask = self.mask_data["masks"]
        outline = self.mask_data["outlines"]

        if mask.ndim == 3:
            mask = np.take(mask, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        if outline.ndim == 3:
            outline = np.take(outline, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        #Create RGBA mask
        image_mask = np.zeros((mask.shape[0], mask.shape[1], 4), dtype=np.uint8)
        r, g, b = self.mask_color
        image_mask[mask != 0] = (r, g, b, self.opacity)
        r, g, b = self.outline_color
        image_mask[outline != 0] = (r, g, b, 255)
        self.image_array = mask

        #Update mask item in the scene
        height, width, _ = image_mask.shape
        qimage = QImage(image_mask.data, width, height, QImage.Format_RGBA8888)
        pixmap = QPixmap.fromImage(qimage)
        if self.mask_item:
            self.scene.removeItem(self.mask_item)
        self.mask_item = self.scene.addPixmap(pixmap)
        self.mask_item.setTransformationMode(Qt.SmoothTransformation)
        self.mask_item.setZValue(1)
        self.mask_item.setVisible(self.mask_show)

        self.scene.setSceneRect(0, 0, width, height)
        self.fitInView(self.sceneRect(), Qt.KeepAspectRatio)

    def load_image_to_scene(self):
        """
        Load the main background image into the scene.
        """
        pixmap = QPixmap(self.adjusted_image_path)

        if self.background_item:
            self.scene.removeItem(self.background_item)
        self.background_item = self.scene.addPixmap(pixmap)
        self.background_item.setZValue(-1)

        self.scene.setSceneRect(0, 0, pixmap.width(), pixmap.height())
        self.fitInView(self.sceneRect(), Qt.KeepAspectRatio)

    def get_npy_of_mask(self):
        mask_path = self.mask_paths[self.image_id][self.bf_channel]
        mask_data = np.load(mask_path, allow_pickle=True).item()

        mask = mask_data["masks"]
        outline = mask_data["outlines"]

        return mask, outline

    def add_drawn_cell(self):
        """
        Adds the new drawn cell to the mask.npy based on current LineItems and updates the scene with the new mask.
        """
        #gets the pixels that build the lines of the drawn cell
        line_pixels = set()
        for item in self.scene.items():
            if isinstance(item, QGraphicsLineItem):
                line = item.line()
                pixels = bresenham_line(line.p1(), line.p2()) #Calculates the pixels along the line
                line_pixels.update(pixels)

        #get the current mask and outline npy arrays
        mask_path = self.mask_paths[self.image_id][self.bf_channel]
        self.mask_data = np.load(mask_path, allow_pickle=True).item()
        mask = self.mask_data["masks"]
        outline = self.mask_data["outlines"]

        if mask.ndim == 3:
            mask = np.take(mask, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        if outline.ndim == 3:
            outline = np.take(outline, int(self.slice_id if self.slice_id is not None else 0), axis=0)

        #Save current state before drawing for undo
        old_state = (mask.copy(), outline.copy())

        free_id = search_free_id(mask, outline)  #search for the next free id in mask and outline

        #add the outline of the new mask (only the parts which not overlap with already existing cells) to outline npy array and fill the complete outline to new_cell_outline to calculate inner pixels
        new_cell_outline = np.zeros_like(outline, dtype=np.uint8)
        for x, y in line_pixels:
            if 0 <= x < outline.shape[1] and 0 <= y < outline.shape[0]:
                new_cell_outline[y, x] = 1
                if outline[y, x] == 0 and mask[y, x] == 0:
                    outline[y, x] = free_id

        #Traces the outline of the new cell and fills the mask based on the outline
        contour = trace_contour(new_cell_outline)
        new_mask = fill_polygon_from_outline(contour, mask.shape) #gets the inner pixels of the new cell
        mask[(new_mask == 1) &(mask == 0) & (outline == 0)] = free_id #adds them to the npy if they not overlap with the already existing cells

        #search if inline pixels (mask) have no outline, if the pixel have no outline neighbor make them to outline and delete them from mask
        new_border_pixels = find_border_pixels(mask,outline,free_id)
        for y, x in new_border_pixels:
            if 0 <= x < outline.shape[1] and 0 <= y < outline.shape[0]:
                mask[y, x] = 0
                outline[y, x] = free_id

        mask_3d = None
        outline_3d = None
        if self.slice_id is not None:
            mask_3d = self.mask_data["masks"]
            outline_3d = self.mask_data["outlines"]

            if mask_3d.ndim == 3:
                mask_3d[int(self.slice_id),:,:] = mask

            if outline_3d.ndim == 3:
                outline_3d[int(self.slice_id),:,:] = outline

        #Save new state after drawing
        np.save(mask_path, {"masks": mask if self.slice_id is None else mask_3d, "outlines": outline if self.slice_id is None else outline_3d}, allow_pickle=True)
        new_state = (mask.copy(), outline.copy())
        self.load_mask_to_scene()
        self.conn.send("update_mask")
        #Save the state transition in undo history
        self.cell_history.append((old_state, new_state))
        self.restoreAvailabilityChanged.emit(len(self.cell_history) > 0)
        self.redoAvailabilityChanged.emit(len(self.redo_history) > 0)
        #Delete the LineItems (the lines that the user have drawn), because the cell now exists through the mask
        for item in list(self.scene.items()):
            if isinstance(item, QGraphicsLineItem):
                self.scene.removeItem(item)



import json
import os
from pathlib import Path
from typing import List, Dict, Any

from PIL.ImageChops import offset
from jsonschema import validate, ValidationError

def load_schema(schema_path: str) -> dict:
    import json
    with open(schema_path, "r", encoding="utf-8") as f:
        return json.load(f)

IGNORED_KEYS = {"view", "position"}

def get_major_dict(data):
    """Deletes the unimportant data for change comparison"""
    if isinstance(data, dict):
        return {k: get_major_dict(v) for k, v in data.items() if k not in IGNORED_KEYS}
    elif isinstance(data, list):
        return [get_major_dict(x) for x in data]
    else:
        return data


class PipelineStorage:
    """
    Handels loading and saving of Pipelines.
    """
    def __init__(self,pipeline_gui):
        self.schema_name = "csp.schema.json"
        self.project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
        self.schema_directory = os.path.join(self.project_root, self.schema_name)
        self.pipeline_gui = pipeline_gui
        self.version = "csp-1.0"
        self.schema = load_schema(schema_path=self.schema_directory)

    def save_as_pipeline(self,file_path:str= ""):
        """
        Saves the pipeline as a JSON file at a specified path.
        """
        pipeline_dict = self.generate_pipline_dict()

        self.pipeline_gui.pipeline_directory = Path(file_path).parent
        self.pipeline_gui.pipeline_dict = pipeline_dict
        self.pipeline_gui.pipeline_name = Path(file_path).stem

        try:
            validate(instance=pipeline_dict, schema=self.schema)
        except ValidationError as e:
            raise ValueError(f"Pipeline json doesn't match with: {e.message}")

        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(pipeline_dict, f, indent=2, ensure_ascii=False)

    def save_pipeline(self):
        """
        Saves the pipeline as a JSON file.
        """
        pipeline_dict = self.generate_pipline_dict()

        self.pipeline_gui.pipeline_dict = pipeline_dict
        from pathlib import Path

        file_path = Path(self.pipeline_gui.pipeline_directory) / f"{self.pipeline_gui.pipeline_name}.csp"

        try:
            validate(instance=pipeline_dict, schema=self.schema)
        except ValidationError as e:
            raise ValueError(f"Pipeline json doesn't match with: {e.message}")

        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(pipeline_dict, f, indent=2, ensure_ascii=False)

        return file_path

    def generate_pipline_dict(self,without_view:bool=False):
        """
        Generate a dictionary(dict) that represents the pipeline from a PipelineGUI instance.
        """
        modules: List[Dict[str, Any]] = []
        pipes: List[Dict[str, Any]] = []
        if not without_view:
            offset_x, offset_y, scale = self.pipeline_gui.interactive_view.get_transformation_data()
            view = {"offset_x": offset_x, "offset_y": offset_y,"scale": scale}
        else:
            view = {"offset_x": 0, "offset_y": 0, "scale": 1}

        for module in self.pipeline_gui.modules.values():
            modules.append(module.to_dict())

        for pipe_list in self.pipeline_gui.pipeline.pipes_in.values():
            for pipe in pipe_list:
                pipes.append(pipe.to_dict())

        pipeline_dict = {
            "version": self.version,
            "version": self.version,
            "modules": modules,
            "pipes": pipes,
            "view": view
        }

        return pipeline_dict

    def check_saved(self):
        """
        Checks if the pipeline is still saved.
        Ignores module positions and view.
        """
        if len(self.pipeline_gui.modules) == 0:
            return True

        if self.pipeline_gui.pipeline_dict == {}:
            return False

        new_pipeline_dict = get_major_dict(self.generate_pipline_dict(without_view=True))
        old_pipeline_dict = get_major_dict(self.pipeline_gui.pipeline_dict)
        return new_pipeline_dict==old_pipeline_dict


    def load_pipeline(self,file_path: str):
        """
        Loads the pipeline into the PipelineGui from the file located at file_path.
        """
        filename = Path(file_path)
        pipeline_dict = {}

        if filename.suffix != ".csp":
            raise ValueError(f"The chosen file has not the right suffix: '{filename.suffix}' expected: '.csp'")

        if filename.exists():
            with filename.open("r", encoding="utf-8") as f:
                pipeline_dict = json.load(f)
        else:
            raise FileNotFoundError("Pipeline json doesn't exist")

        try:
            validate(instance=pipeline_dict, schema=self.schema)
        except ValidationError as e:
            raise ValueError(f"Pipeline json doesn't match with schema: {self.schema_directory}")

        self.pipeline_gui.pipeline_directory = filename.parent
        self.pipeline_gui.pipeline_name = filename.stem
        self.pipeline_gui.pipeline_dict = pipeline_dict




import flet as ft
from cellpose import models, train, io
import os

from cellsepi.frontend.main_window.gui_directory import format_directory_path, copy_to_clipboard


class Training(ft.Container):

    def __init__(self, gui):
        super().__init__()
        self.gui = gui
        self.text = ft.Text("Go To Training")
        self.button_event = ft.PopupMenuItem(
            content=ft.Row(
                [
                    ft.Icon(ft.Icons.EXIT_TO_APP),
                    self.text,
                ]
            ),
            on_click=lambda e: self.change_environment(e),
        )
        self.switch_icon = ft.Icon(ft.Icons.MODEL_TRAINING)
        self.button_training_environment_menu = ft.PopupMenuButton(
            items=[self.button_event],
            content=self.switch_icon,
            tooltip="Training",
            on_open=lambda _: self.text.update(),
        )
        self.content = self.button_training_environment_menu
        self.padding = 10
        self.alignment = ft.alignment.top_right
        self.start_button = ft.ElevatedButton(
            text="Start",
            icon=ft.Icons.PLAY_CIRCLE,
            tooltip="Start the training epochs",
            disabled=True,
            on_click=self.start_training,
        )

        self.model = "nuclei"
        self.batch_size = 100
        self.epochs = 100
        self.learning_rate = 0.001
        self.pre_trained = None
        self.diameter_default = True
        self.diameter = self.gui.average_diameter.get_avg_diameter()
        self.weight = 1e-4  # standard value for the weight
        self.model_name = "new_model"
        self.re_train_model_name = None
        self.color = ft.Colors.BLUE_400
        self.progress_bar_text = ft.Text("")

        project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        self.model_directory = os.path.join(project_root, "models")

        # Changed from TextField to Dropdown for model type selection
        self.model_dropdown = ft.Dropdown(
            label="Model Type",
            value="nuclei",
            options=[
                ft.dropdown.Option("nuclei"),
                ft.dropdown.Option("cyto"),
                ft.dropdown.Option("cyto2"),
                ft.dropdown.Option("cyto3")
            ],border_color=ft.Colors.BLUE_400,
            on_change=lambda e: self.changed_input("modeltype", e),expand=True,
        )
        self.re_train_model = ft.Checkbox(value=False, label="Retrain Model",on_change=lambda e: self.change_re_train_model())

        # the following methods are called when clicking on the corresponding button
        def pick_model_result(e: ft.FilePickerResultEvent):
            """
            The result of the file selection is handled.

            Arguments:
                e (ft.FilePickerResultEvent): the result of the file picker event, i.e. the chosen file
            """
            if e.files is None:
                #case: no model selected
                pass
            elif e.files[0].path is not None:
                self.gui.csp.re_train_model_path = e.files[0].path
                self.field_model_name.value = e.files[0].name
                self.re_train_model_name = e.files[0].name
                self.field_model_name.color = ft.Colors.BLUE_400
                self.gui.page.update()

        pick_model_dialog = ft.FilePicker(on_result=pick_model_result)
        self.gui.page.overlay.extend([pick_model_dialog])

        self.re_train_model_chooser = ft.IconButton(
                icon=ft.Icons.UPLOAD_FILE,
                tooltip="Choose model to retrain",
                on_click=lambda _: pick_model_dialog.pick_files(allow_multiple=False,
                                                                initial_directory=self.model_directory),disabled=True
            )
        self.field_model_name = ft.TextField(label="Model Name", value=self.model_name, border_color=self.color,on_change=lambda e: self.changed_input("model_name", e))
        self.model_stack = ft.Stack([self.field_model_name, self.re_train_model_chooser],alignment=ft.alignment.top_right)
        self.field_model = ft.Row([self.model_dropdown, self.model_stack])
        # New field for custom model input, visible only if "custom" is selected
        self.field_custom_model = ft.TextField(label="Custom Model", value="", border_color=self.color, visible=False,
                                               on_change=lambda e: self.changed_input("custom_model", e))

        self.field_batch = ft.TextField(label="Batch Size", value=self.batch_size, border_color=self.color,
                                        on_change=lambda e: self.changed_input("batch_size", e))
        self.field_epoch = ft.TextField(label="Epochs", value=self.epochs, border_color=self.color,
                                        on_change=lambda e: self.changed_input("epochs", e))
        self.field_lr = ft.TextField(label="Learning Rate", value=self.learning_rate, border_color=self.color,
                                     on_change=lambda e: self.changed_input("learning_rate", e))
        self.field_diameter = ft.TextField(label="Diameter", value=self.diameter, border_color=self.color,
                                           on_change=lambda e: self.changed_input("diameter", e))
        self.field_weights = ft.TextField(label="Weight Decay", value=self.weight, border_color=self.color,
                                          on_change=lambda e: self.changed_input("weight", e))
        self.field_directory = ft.TextField(label="Directory", value=format_directory_path(self.model_directory,60), border_color=self.color,
                                            read_only=True,disabled=True)

        self.directory_stack = ft.Stack([self.field_directory,ft.Container(
                            content=ft.Container(
                                content=ft.IconButton(
                                    icon=ft.Icons.COPY,
                                    tooltip="Copy to clipboard",
                                    on_click=lambda e: copy_to_clipboard(self.gui.page,self.model_directory,"Model directory")
                                ),
                                alignment=ft.alignment.top_right,
                            )
                        )])
        self.progress_ring = ft.ProgressRing(visible=False)
        self.train_loss = None
        self.test_loss = None

    def change_re_train_model(self):
        """
        Choosing a model to retrain.
        """
        self.field_model_name.disabled = self.re_train_model.value
        if self.re_train_model.value is True:
            self.re_train_model_chooser.disabled = False
            self.field_diameter.disabled = True
            self.field_diameter.value = None
            self.model_dropdown.visible = False
            if self.re_train_model_name is not None:
                self.field_model_name.value = self.re_train_model_name
                self.field_model_name.color = ft.Colors.BLUE_400
            else:
                self.field_model_name.value = None
        else:
            self.re_train_model_chooser.disabled = True
            self.field_diameter.disabled = False
            self.field_diameter.value = self.diameter
            self.field_model_name.color = None
            self.field_model_name.value = self.model_name
            self.model_dropdown.visible = True
        self.gui.page.update()

    def go_to_training_environment(self, e):
        # delete the content of the page and reset the reference to the page (reference get sometimes lost)
        self.gui.ref_training_environment.current.visible = True
        self.gui.ref_gallery_environment.current.visible = True
        self.gui.ref_builder_environment.current.visible = False
        self.gui.ref_seg_environment.current.visible = False
        self.page.title = "CellSePi"
        self.gui.page.update()
        self.text.value = "Exit Training"
        self.gui.ex_mode.text.value = "Go To Expert Mode"

    def add_parameter_container(self):
        return ft.Container(
            ft.Column(
                [self.field_model,self.re_train_model, self.field_custom_model, self.field_batch, self.field_epoch, self.field_weights,
                 self.field_lr, self.field_diameter, self.directory_stack
                 ]
            ), padding=10,
        )

    def changed_input(self, field, e):
        """
        Changing the value of one of the parameters for training.
        Arguments:
            field: the parameter to change
            e = the change event
        """
        updated_value = e.control.value

        if field == "modeltype":
            self.model = updated_value
            self.field_model.value = updated_value
            if updated_value == "custom":
                self.field_custom_model.visible = True
            else:
                self.field_custom_model.visible = False
        elif field == "custom_model":
            self.model = updated_value
            self.field_custom_model.value = updated_value
        elif field == "batch_size":
            self.batch_size = int(updated_value)
            self.field_batch.value = updated_value
        elif field == "epochs":
            self.epochs = int(updated_value)
            self.field_epoch.value = updated_value
        elif field == "learning_rate":
            self.learning_rate = float(updated_value)
            self.field_lr.value = updated_value
        elif field == "pre_trained":
            self.pre_trained = updated_value
            self.field_trained.value = updated_value
        elif field == "weight":
            self.weight = float(updated_value)
            self.field_weights.value = updated_value
        elif field == "model_name":
            self.model_name = updated_value
        else:
            self.diameter_default = False
            self.diameter = float(updated_value)
            self.field_diameter.value = updated_value

        self.gui.page.update()

    def change_environment(self, e):
        if self.text.value == "Go To Training":
            self.go_to_training_environment(e)
        else:
            self.gui.ref_training_environment.current.visible = False
            self.gui.ref_seg_environment.current.visible = True
            self.gui.page.update()
            self.text.value = "Go To Training"

    def create_training_card(self):
        """
        This method creates a card for the GUI, which contains the progress bar and several buttons for
         controlling the run of the training.

        Returns:
            training_card (ft.Card): the card containing all the elements needed to run the training
        """

        # progress bar, which is updated throughout the training periods

        text = ft.Text("Training")
        title = ft.ListTile(
            leading=ft.Icon(name=ft.Icons.HUB_OUTLINED),
            title=text,
        )
        pick_model_row = ft.Row(
            [
                ft.Container(content=ft.Row([self.progress_ring, self.progress_bar_text]),padding=5),
                ft.Container(
                    content=ft.Row([self.start_button]))
            ], alignment=ft.MainAxisAlignment.SPACE_BETWEEN
        )
        test_container = ft.Container(
            content=ft.Column(
                [title,
                 pick_model_row,
                 ]
            )
        )

        progress_card = ft.Card(
            content=ft.Container(
                content=ft.Stack(
                    [test_container]
                ),
                padding=10
            ),
        )

        return progress_card

    def start_training(self, e):
        """
        This method starts the training process with the selected parameters and model.
        """
        self.start_button.disabled = True
        self.gui.directory.disable_path_choosing()
        self.progress_ring.visible = True
        self.progress_bar_text.value = ""
        self.disable_switch_environment()
        self.gui.page.update()

        # checks if the right model type was selected
        if self.re_train_model.value and self.re_train_model_name is None:
            self.page.open(ft.SnackBar(
                ft.Text(f"The model you inserted is not a retrained model!")))
            self.gui.directory.enable_path_choosing()
            self.start_button.disabled = False
            self.progress_ring.visible = False
            self.progress_bar_text.value = ""
            self.enable_switch_environment()
            self.page.update()
            return
        self.gui.csp.training_running = True
        self.gui.queue.put("delete_mask")
        try:
            mask_filter = f"{self.gui.csp.current_mask_suffix}.npy"

            # loads the mask files out of the directory to start training
            output = io.load_train_test_data(train_dir=str(self.gui.csp.working_directory),
                                             mask_filter=mask_filter,
                                             look_one_level_down=False)
            images, labels, image_names, test_images, test_labels, image_names_test = output

        except Exception as e:
            self.page.open(ft.SnackBar(
                ft.Text(f"Something went wrong while gather training data: {str(e)}")))
            self.gui.directory.enable_path_choosing()
            self.start_button.disabled = False
            self.progress_ring.visible = False
            self.progress_bar_text.value = ""
            self.enable_switch_environment()
            self.page.update()
            self.gui.csp.training_running = False
            if self.gui.training_event is not None:
                self.gui.training_event.set()
            return
        if len(images) == 0 or len(labels) == 0:
            self.page.open(ft.SnackBar(
                ft.Text(f"You need images and suitable masks to train a model!")))
            self.gui.directory.enable_path_choosing()
            self.start_button.disabled = False
            self.progress_ring.visible = False
            self.progress_bar_text.value = ""
            self.enable_switch_environment()
            self.page.update()
            self.gui.csp.training_running = False
            if self.gui.training_event is not None:
                self.gui.training_event.set()
            return
        try:
            # initializing variables, who differ if pretrained or not (Initialized with not pretrained)
            sgd_value = False
            model_name = self.model_name
            model = models.CellposeModel(model_type=self.model, diam_mean=self.diameter)
            if self.re_train_model.value:
                sgd_value = True
                model_name = self.re_train_model_name
                model = models.CellposeModel(model_type=None, pretrained_model=self.gui.csp.re_train_model_path)

            # start the training epochs
            train.train_seg(model.net,
                            train_data=images, train_labels=labels,
                            channels=[1, 2], normalize=True,
                            test_data=test_images, test_labels=test_labels,
                            weight_decay=self.weight, SGD=sgd_value, learning_rate=self.learning_rate,
                            n_epochs=self.epochs, model_name=model_name,
                            save_path=os.path.dirname(self.model_directory))
            self.progress_bar_text.value = "Finished Training"

        except Exception as e:
            self.page.open(ft.SnackBar(
                ft.Text(f"Something went wrong while training: {str(e)}")))
            self.progress_bar_text.value = ""
            self.page.update()
        self.gui.directory.enable_path_choosing()
        self.start_button.disabled = False
        self.progress_ring.visible = False
        self.enable_switch_environment()
        self.page.update()
        self.gui.csp.training_running = False
        if self.gui.training_event is not None:
            self.gui.training_event.set()

    def disable_switch_environment(self):
        self.switch_icon.color = ft.Colors.GREY_400
        self.button_training_environment_menu.disabled = True
        self.button_training_environment_menu.update()
        self.gui.ex_mode.switch_icon.color = ft.Colors.GREY_400
        self.gui.ex_mode.button_expert_environment_menu.disabled = True
        self.gui.ex_mode.button_expert_environment_menu.update()

    def enable_switch_environment(self):
        self.switch_icon.color = None
        self.button_training_environment_menu.disabled = False
        self.button_training_environment_menu.update()
        self.gui.ex_mode.switch_icon.color = None
        self.gui.ex_mode.button_expert_environment_menu.disabled = False
        self.gui.ex_mode.button_expert_environment_menu.update()



import flet as ft

from cellsepi.frontend.main_window.gui_colors import ColorSelection, ColorOpacity


class Options(ft.Container):
    """
    Class which handles the options in the right up corner in the GUI.
    """
    def __init__(self, gui):
        super().__init__()
        self.page = gui.page
        self.gui = gui
        self.dark_light_text = ft.Text("Light Theme")
        self.dark_light_icon = ft.IconButton(
            icon=ft.Icons.BRIGHTNESS_2_OUTLINED,
            icon_color=None,
            on_click=self.theme_changed,
        )
        self.color_selection = ColorSelection(gui)
        self.color_opacity= ColorOpacity(gui)
        self.menu_button = ft.PopupMenuButton(
            items=self.create_appbar_items(),
            content=ft.Icon(ft.Icons.MENU),
            tooltip="Options",
            on_open=self.check_current_theme,
        )
        self.content = self.menu_button
        self.padding = 10
        self.alignment = ft.alignment.top_right

    async def theme_changed(self, e):
        """
        Changes the theme of the page to the opposite of the current selected theme.
        """
        if self.page.theme_mode == ft.ThemeMode.LIGHT:
            self.page.theme_mode = ft.ThemeMode.DARK
            self.dark_light_text.value = "Dark Theme"
            self.dark_light_icon.icon = ft.Icons.BRIGHTNESS_HIGH
        else:
            self.page.theme_mode = ft.ThemeMode.LIGHT
            self.dark_light_text.value = "Light Theme"
            self.dark_light_icon.icon = ft.Icons.BRIGHTNESS_2_OUTLINED
        self.page.update()

    def check_current_theme(self,e):
        """
        Checks what the current theme is.
        """
        if self.page.theme_mode == ft.ThemeMode.SYSTEM:
            if self.page.platform_brightness == ft.Brightness.LIGHT:
                self.dark_light_text.value = "Light Theme"
                self.dark_light_icon.icon = ft.Icons.BRIGHTNESS_2_OUTLINED
            else:
                self.dark_light_text.value = "Dark Theme"
                self.dark_light_icon.icon = ft.Icons.BRIGHTNESS_HIGH
        else:
            if self.page.theme_mode == ft.ThemeMode.LIGHT:
                self.dark_light_text.value = "Light Theme"
                self.dark_light_icon.icon = ft.Icons.BRIGHTNESS_2_OUTLINED
            else:
                self.dark_light_text.value = "Dark Theme"
                self.dark_light_icon.icon = ft.Icons.BRIGHTNESS_HIGH
        self.page.update()

    def create_appbar_items(self):
        """
        Creates the appbar items that will be displayed in the GUI when the option button is clicked.
        """
        return [
            ft.PopupMenuItem(
                content=ft.Row([self.dark_light_icon, self.dark_light_text]),
                on_click=self.theme_changed,
            ),
            ft.PopupMenuItem(
                content=ft.Row([self.color_selection.color_icon_mask, ft.Text("Mask Color")]),
                on_click=self.color_selection.open_color_picker_mask,
            ),
            ft.PopupMenuItem(
                content=ft.Row([self.color_selection.color_icon_outline, ft.Text("Outline Color")]),
                on_click=self.color_selection.open_color_picker_outline,
            ),
            ft.PopupMenuItem(
                content=ft.Container(
                    content=ft.Column(
                [
                            ft.Container(
                                content=self.color_opacity.text,
                                padding=ft.padding.only(bottom=-10)
                            ),
                            ft.Container(
                                content=self.color_opacity.slider,
                                padding=ft.padding.only(bottom=-8)
                            ),
                        ],
                        spacing=0,
                        alignment=ft.MainAxisAlignment.CENTER,
                        horizontal_alignment=ft.CrossAxisAlignment.CENTER
                    ),
                    padding=ft.padding.all(0),
                ),
            )
        ]



import multiprocessing
import os
import threading
from concurrent.futures import ThreadPoolExecutor

import cv2
import pandas as pd

import torch
import torchvision.transforms as T
import numpy as np
from PIL import Image
from cellpose import models, io
from cellpose.io import imread
from scipy.ndimage import binary_erosion
from tifffile import tifffile
from torchvision.models.detection import maskrcnn_resnet50_fpn
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor

from cellsepi.backend.drawing_window.drawing_util import trace_contour
from cellsepi.backend.main_window.data_util import load_image_to_numpy
from cellsepi.backend.main_window.expert_mode.event_manager import EventManager
from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent
from cellsepi.backend.main_window.notifier import Notifier
from cellsepi.frontend.main_window.gui_mask import reset_mask, handle_mask_update


class BatchImageSegmentation(Notifier):
    """
    This class handles the segmentation of the images.
    """
    def __init__(self,
                 segmentation = None,
                 gui = None,
                 device = "cpu",segmentation_channel:str = "", diameter:float = 125.0,suffix:str = "_seg"):
        if gui is not None:
            super().__init__()
            self.segmentation = segmentation
            self.gui = gui
            self.segmentation_channel = self.gui.csp.config.get_bf_channel()
            self.diameter = self.gui.csp.config.get_diameter()
            self.suffix = self.gui.csp.current_mask_suffix
        else:
            self.segmentation_channel = segmentation_channel
            self.diameter = diameter
            self.suffix = suffix

        self.masks_backup = {}
        self.device = device
        self.prev_masks_exist = False
        self.num_seg_images = 0
        self.cancel_now = False
        self.pause_now = False
        self.resume_now = False
        self.executor = None
        self.progress_lock = threading.Lock()
        self.progress = 0

    def _is_cellpose_model(self, model_path):
        try:
            from cellpose import models
            _ = models.CellposeModel(pretrained_model=model_path, device=torch.device(self.device))
            return True
        except Exception:
            return False

    def get_contour_from_labeled_mask(label_mask):
        outlines = np.zeros_like(label_mask, dtype=np.uint8)
        obj_ids = np.unique(label_mask)
        obj_ids = obj_ids[obj_ids != 0]

        for obj_id in obj_ids:
            mask = label_mask == obj_id
            eroded = binary_erosion(mask)
            outline = mask & (~eroded)
            outlines[outline] = 255

        return outlines

    def masks_to_label_mask(self, masks):
        N, H, W = masks.shape
        label_mask = np.zeros((H, W), dtype=np.int32)

        for i in range(N):
            mask = masks[i].astype(bool)
            label_mask[mask] = i + 1

        return label_mask

    def backup_masks(self):
        """
        This method creates a backup of the previously generated masks.
        """
        self.prev_masks_exist = False

        for image_id, channels in self.gui.csp.mask_paths.items():
            for segmentation_channel, path in channels.items():
                if segmentation_channel == self.segmentation_channel:
                    if os.path.exists(path):
                        self.masks_backup[image_id] = {}
                        self.prev_masks_exist = True
                        mask = np.load(path,allow_pickle=True)
                        self.masks_backup[image_id][segmentation_channel] = mask

        if self.prev_masks_exist:
            for image_id in self.gui.csp.image_paths:
                if image_id not in self.masks_backup:
                    self.masks_backup[image_id] = {}
                    self.masks_backup[image_id][self.segmentation_channel] = None

    def delete_mask(self, path, channels_to_delete, image_id, segmentation_channel):
        if os.path.exists(path):
            channels_to_delete.append((image_id, segmentation_channel))
            if image_id == self.gui.csp.image_id:
                if self.segmentation_channel == segmentation_channel:
                    self.gui.switch_mask.value = False  # sets the mask switch to False because there is no longer a mask
                    self.gui.canvas.container_mask.visible = False  # and sets the mask picture invisible because it is no longer valid
                    reset_mask(self.gui, image_id,segmentation_channel)
                    self.gui.page.update()
            if image_id == self.gui.csp.window_image_id:
               if segmentation_channel == self.gui.csp.window_bf_channel:
                   self.gui.queue.put("delete_mask")  # sends the info that the current image is deleted to the drawing window
            os.remove(path)

    def restore_backup(self):
        """
        This method restores the previously generated masks and deletes the old ones.
        """
        channels_to_delete = []
        if self.prev_masks_exist:
            for image_id, channels in self.masks_backup.items():
                if image_id not in self.gui.csp.mask_paths:
                    self.gui.csp.mask_paths[image_id] = {}
                for segmentation_channel, mask in channels.items():
                    if mask is not None:
                        backup_path = self.gui.csp.mask_paths[image_id].get(segmentation_channel)
                        if backup_path:
                            np.save(backup_path, mask)
                            if image_id == self.gui.csp.window_image_id:
                                if segmentation_channel == self.gui.csp.window_bf_channel:
                                    self.gui.queue.put("refresh_mask")  # refreshes if the backup is the current selected image and the mask is the same channel
                            if image_id == self.gui.csp.image_id and segmentation_channel == self.gui.csp.config.get_bf_channel(): #refreshes or delete the current generated mask
                                handle_mask_update(self.gui)
                            else:
                                reset_mask(self.gui, image_id,segmentation_channel)
                    else:
                        if segmentation_channel in self.gui.csp.mask_paths[image_id]:
                            path = self.gui.csp.mask_paths[image_id][segmentation_channel]
                            self.delete_mask(path, channels_to_delete, image_id, segmentation_channel)
        else:  # case where no masks for this bf_channel existed before
            for image_id, channels in self.gui.csp.mask_paths.items():
                for segmentation_channel, path in channels.items():
                    if segmentation_channel == self.segmentation_channel:
                        self.delete_mask(path, channels_to_delete, image_id, segmentation_channel)

        for image_id, segmentation_channel in channels_to_delete:
            del self.gui.csp.mask_paths[image_id][segmentation_channel]

        for image_id, channel in channels_to_delete:
            del self.gui.csp.mask_paths[image_id]


    # the following methods handle the different actions and handle accordingly
    def cancel_action(self):
        self.cancel_now = True
        if self.executor is not None:
            self.executor.shutdown(wait=True)

    def pause_action(self):
        self.pause_now = True
        if self.executor is not None:
            self.executor.shutdown(wait=True)

    def resume_action(self):
        self.resume_now = True

    def run(self, event_manager: EventManager= None, image_paths = None, mask_paths=None, model_path = None):
        """
        Applies the segmentation model to every image and stores the resulting masks.
        """
        if event_manager is None:
            if self.num_seg_images == 0:  # shouldn't backup again, if it was paused and now resuming
                self.backup_masks()
                self.segmentation_channel = self.gui.csp.config.get_bf_channel()
                self.diameter = self.gui.csp.config.get_diameter()
                self.suffix = self.gui.csp.current_mask_suffix
            if self.cancel_now:
                self.cancel_now = False
                self.restore_backup()
                self.num_seg_images = 0
                return
            elif self.pause_now:
                self.pause_now = False
                return
            elif self.resume_now:
                self.resume_now = False
                self.segmentation.is_resuming()


        if event_manager is None:
            self._call_start_listeners()
        if event_manager is None:
            image_paths = self.gui.csp.image_paths

        segmentation_channel = self.segmentation_channel
        diameter = self.diameter
        suffix = self.suffix

        n_images = len(image_paths)

        if event_manager is None:
            segmentation_model = self.gui.csp.model_path
        else:
            segmentation_model = model_path
            event_manager.notify(ProgressEvent(0, f"Segmenting Images: 0/{n_images}"))

        device = torch.device(self.device)  # converts string to device object

        io.logger_setup()  # configures logging system for Cellpose

        if self._is_cellpose_model(segmentation_model):
            model_type = 'cellpose'
            model = models.CellposeModel(device=device, pretrained_model=segmentation_model)
        else:
            model_type = 'pytorch'
            model = maskrcnn_resnet50_fpn(weights="DEFAULT")
            in_features = model.roi_heads.box_predictor.cls_score.in_features
            model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes=2)
            in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
            model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask, 256,num_classes=2)

            model.load_state_dict(torch.load(segmentation_model, map_location=self.device))
            model.to(self.device)
            model.eval()

        start_index = self.num_seg_images
        for iN, image_id in enumerate(list(image_paths)[start_index:], start=start_index):
            if segmentation_channel in image_paths[image_id] and os.path.isfile(image_paths[image_id][segmentation_channel]):
                if event_manager is None:
                    if self.cancel_now:
                        self.cancel_now = False
                        self.restore_backup()
                        self.num_seg_images = 0
                        return
                    elif self.pause_now:
                        self.pause_now = False
                        return
                    elif self.resume_now:
                        self.resume_now = False
                        self.segmentation.is_resuming()

                image_path = image_paths[image_id][segmentation_channel]
                image = tifffile.imread(image_path)

                # Normalization
                image = image.astype(np.float32)
                min_val = np.min(image)
                max_val = np.max(image)
                if (max_val - min_val) > 0:
                    image = (image - min_val) / (max_val - min_val)
                else:
                    image = np.zeros_like(image)

                # model evaluates image
                if model_type == 'cellpose':
                    if image.ndim == 3: #x,y,z dimensions
                        res = model.eval(image, diameter=diameter, channels=[0, 0],z_axis=2,do_3D=False)
                    else:
                        res = model.eval(image, diameter=diameter, channels=[0, 0])
                    mask, flow, style = res[:3]

                elif model_type == 'pytorch':
                    # PyTorch models expect tensor with shape [C, H, W], normalized
                    if image.ndim == 2:  # grayscale
                        image = np.stack([image] * 3, axis=-1)
                    elif image.shape[2] == 1:
                        image = np.concatenate([image] * 3, axis=-1)

                    pil_img = Image.fromarray((image * 255).astype(np.uint8))
                    transform = T.ToTensor()
                    img_tensor = transform(pil_img).to(device)

                    with torch.no_grad():
                        prediction = model([img_tensor])[0]

                    if 'masks' not in prediction or len(prediction['masks']) == 0:
                        mask = np.zeros_like(image[..., 0], dtype=np.uint16)
                    else:
                        masks = prediction['masks'] > 0.5
                        mask = masks.squeeze(1).cpu().numpy()

                        mask = self.masks_to_label_mask(mask)

                    flow, style = None, None

                # Generate the output filename directly using the suffix attribute
                directory, filename = os.path.split(image_path)
                name, _ = os.path.splitext(filename)
                new_filename = f"{name}{suffix}.npy"
                new_path = os.path.join(directory, new_filename)

                default_suffix_path = os.path.splitext(image_path)[0] + '_seg.npy'

                backup_path = None
                if default_suffix_path != new_path:
                    if os.path.exists(default_suffix_path):
                        backup_path = default_suffix_path + '.backup'
                        if os.path.exists(backup_path):
                            os.remove(backup_path)
                        os.rename(default_suffix_path, backup_path)
                # Save the segmentation results directly with the default name first
                if model_type == 'cellpose':
                    io.masks_flows_to_seg([image], [mask], [flow], [image_path])
                else:
                    H, W = image.shape[:2]
                    flow0 = np.zeros((H, W, 3), dtype=np.uint8)
                    flow1 = np.zeros((2, H, W), dtype=np.float32)
                    flow2 = np.zeros((H, W), dtype=np.float32)
                    dummy_flow = [flow0, flow1, flow2]
                    io.masks_flows_to_seg([image], [mask], [dummy_flow], [image_path])

                if default_suffix_path != new_path:
                    if os.path.exists(default_suffix_path):
                        if os.path.exists(new_path):
                            os.remove(new_path)
                        os.rename(default_suffix_path, new_path)
                        if backup_path is not None:
                            os.rename(backup_path, default_suffix_path)
                if event_manager is None:
                    if image_id not in self.gui.csp.mask_paths:
                            self.gui.csp.mask_paths[image_id] = {}
                else:
                    if image_id not in mask_paths:
                        mask_paths[image_id] = {}

                if event_manager is None:
                    self.gui.csp.mask_paths[image_id][segmentation_channel] = new_path
                else:
                    mask_paths[image_id][segmentation_channel] = new_path

                percent= round((iN + 1) / n_images * 100)
                progress = str(percent) + " %"
                if event_manager is None:
                    current_image = {"image_id": image_id, "path": image_path}
                    self._call_update_listeners(progress, current_image)
                else:
                    event_manager.notify(ProgressEvent(percent=percent,process=f"Segmenting Images: {iN+1}/{n_images} (Latest Image: {image_id})"))
                self.num_seg_images = self.num_seg_images + 1
                if event_manager is None:
                    self.gui.directory.update_mask_check(image_id)
                    self.gui.diameter_text.value = self.gui.average_diameter.get_avg_diameter()
            else:
                percent= round((iN + 1) / n_images * 100)
                progress = str(percent) + " %"
                if event_manager is None:
                    current_image = {"image_id": image_id, "path": None}
                    self._call_update_listeners(progress, current_image)
                else:
                    event_manager.notify(ProgressEvent(percent=percent,process=f"Segmenting Images: {iN+1}/{n_images} (Latest Image: {image_id})"))
                self.num_seg_images = self.num_seg_images + 1

        if event_manager is None:
            self._call_completion_listeners()
        else:
            event_manager.notify(ProgressEvent(percent=100 ,process=f"All images segmented."))
        # reset variables
        self.num_seg_images = 0

    """def run_parallel(self):
        
        Applies the segmentation model to every image in parallel and stores the resulting masks.
        
        if self.num_seg_images == 0:  # shouldn't backup again, if it was paused and now resuming
            self.backup_masks()
            self.segmentation_channel = self.gui.csp.config.get_bf_channel()
            self.diameter = self.gui.csp.config.get_diameter()
            self.suffix = self.gui.csp.current_mask_suffix
        if self.cancel_now:
            self.cancel_now = False
            self.restore_backup()
            self.num_seg_images = 0
            return
        elif self.pause_now:
            self.pause_now = False
            return
        elif self.resume_now:
            self.resume_now = False
            self.segmentation.is_resuming()

        self._call_start_listeners()
        image_paths = self.gui.csp.image_paths
        segmentation_channel = self.segmentation_channel
        diameter = self.diameter
        suffix = self.suffix

        segmentation_model = self.gui.csp.model_path
        device = self.device
        device = torch.device(device)  # converts string to device object

        io.logger_setup()  # configures logging system for Cellpose
        model = models.CellposeModel(device=device, pretrained_model=segmentation_model)

        start_index = self.num_seg_images
        self.executor = ThreadPoolExecutor(max_workers=4)
        futures = []
        for iN, image_id in enumerate(list(image_paths)[start_index:], start=start_index):
            futures.append(self.executor.submit(
                self.image_segmentation,
                iN, image_id, image_paths, segmentation_channel, diameter, suffix, model
            ))
        for future in futures:
            future.result()
        if self.cancel_now:
            self.cancel_now = False
            self.restore_backup()
            self.num_seg_images = 0
            return
        self._call_completion_listeners()
        # reset variables
        self.num_seg_images = 0
    
    def image_segmentation(self,image_id, image_paths, segmentation_channel, diameter, suffix, model):
        
        Applies the segmentation model to a single image.

        Attributes:
            image_id: identification number of the image
            image_paths: list of image paths
            segmentation_channel: bright field channel
            suffix: suffix to be applied to the mask filename
            model: cellpose model to be used
        
        n_images = len(image_paths)
        if self.cancel_now:
            self.cancel_now = False
            self.restore_backup()
            self.num_seg_images = 0
            return
        elif self.pause_now:
            self.pause_now = False
            return
        elif self.resume_now:
            self.resume_now = False
            self.segmentation.is_resuming()

        if segmentation_channel in image_paths[image_id] and os.path.isfile(image_paths[image_id][segmentation_channel]):
            image_path = image_paths[image_id][segmentation_channel]
            image = imread(image_path)
            # Normalization
            image = image.astype(np.float32)
            min_val = np.min(image)
            max_val = np.max(image)
            if (max_val - min_val) > 0:
                image = (image - min_val) / (max_val - min_val)
            else:
                image = np.zeros_like(image)

            res = model.eval(image, diameter=diameter, channels=[0, 0])
            mask, flow, style = res[:3]
            # Generate the output filename directly using the suffix attribute
            directory, filename = os.path.split(image_path)
            name, _ = os.path.splitext(filename)
            new_filename = f"{name}{suffix}.npy"
            new_path = os.path.join(directory, new_filename)

            default_suffix_path = os.path.splitext(image_path)[0] + '_seg.npy'
            backup_path = None
            if default_suffix_path != new_path:
                if os.path.exists(default_suffix_path):
                    backup_path = default_suffix_path + '.backup'
                    if os.path.exists(backup_path):
                        os.remove(backup_path)
                    os.rename(default_suffix_path, backup_path)
            # Save the segmentation results directly with the default name first
            io.masks_flows_to_seg([image], [mask], [flow], [image_path])
            if default_suffix_path != new_path:
                if os.path.exists(default_suffix_path):
                    if os.path.exists(new_path):
                        os.remove(new_path)
                    os.rename(default_suffix_path, new_path)
                    if backup_path is not None:
                        os.rename(backup_path, default_suffix_path)
            if image_id not in self.gui.csp.mask_paths:
                self.gui.csp.mask_paths[image_id] = {}

            self.gui.csp.mask_paths[image_id][segmentation_channel] = new_path

            with self.progress_lock:
                self.progress += 1
                percent = round(self.progress / n_images * 100)
                progress = str(percent) + "%"
                current_image = {"image_id": image_id, "path": image_path}
                self._call_update_listeners(progress, current_image)
            self.num_seg_images = self.num_seg_images + 1
            self.gui.directory.update_mask_check(image_id)
            self.gui.diameter_text.value = self.gui.average_diameter.get_avg_diameter()
        else:
            with self.progress_lock:
                self.progress += 1
                percent = round(self.progress / n_images * 100)
                progress = str(percent) + "%"
                current_image = {"image_id": image_id, "path": None}
                self._call_update_listeners(progress, current_image)
            self.num_seg_images = self.num_seg_images + 1

        if self.cancel_now:
            self.cancel_now = False
            self.restore_backup()
            self.num_seg_images = 0
            return
        elif self.pause_now:
            self.pause_now = False
            return
        elif self.resume_now:
            self.resume_now = False
            self.segmentation.is_resuming()
    """

class BatchImageReadout(Notifier):

    def __init__(self, image_paths,
                 mask_paths,
                 segmentation_channel,
                 channel_prefix="c",
                 directory=None,module:bool = False):
        if not module:
            super().__init__()

        if directory is None:
            directory = ""

        self.image_paths = image_paths
        self.mask_paths = mask_paths
        self.segmentation_channel = segmentation_channel
        self.channel_prefix = channel_prefix
        self.directory = directory

    def _channel_name(self, channel_id):
        return self.channel_prefix + str(channel_id)

    def run(self,event_manager: EventManager = None):
        image_paths = self.image_paths
        n_images = len(image_paths)

        if event_manager is None:
            self._call_start_listeners()
        else:
            event_manager.notify(ProgressEvent(0,f"Readout Images: 0/{n_images}"))

        mask_paths = self.mask_paths
        segmentation_channel = self.segmentation_channel


        row_entries = []

        for iN, image_id in enumerate(image_paths):

            # 1. Check if Image has Mask in mask_paths
            # 2. Iterate over all channels and skip segmentation channel
            # 3. Get Background and derive
            # 4. For each cell readout fluorescence
            # 5. Store values in a pandas dataframe "readout" (Layout: Image ID | Cell ID | Channels ... | Background)

            if not image_id in mask_paths or not segmentation_channel in mask_paths[image_id]:
                continue
            mask_path = mask_paths[image_id][segmentation_channel]
            mask_data = np.load(mask_path,allow_pickle=True).item()
            mask = mask_data["masks"]

            if mask.ndim == 3:
                mask = np.transpose(mask, (1, 2, 0))

            cell_ids = np.unique(mask)
            if len(cell_ids) == 1:
                if event_manager is None:
                    self._call_update_listeners(**kwargs)
                else:
                    event_manager.notify(ProgressEvent(percent=int((iN + 1) / n_images * 100),
                                                       process=f"Readout Images: {iN + 1}/{n_images} (Latest Image: {image_id})"))
                continue
            cell_ids = cell_ids[1:]

            channels = list(image_paths[image_id])
            n_channels = len(channels)

            cur_row_entries = [None] * len(cell_ids)
            for iX, cell_id in enumerate(cell_ids):
                data_entry = {"image_id": image_id,
                              "cell_id": cell_id}
                for channel_id in channels:
                    channel_name = self._channel_name(channel_id)
                    data_entry[channel_name] = None
                    data_entry[f"background {channel_name}"] = None

                cur_row_entries[iX] = data_entry

            for channel_id in channels:
                image_path = image_paths[image_id][channel_id]
                channel_name = self._channel_name(channel_id)

                np_image = load_image_to_numpy(image_path)
                background_mask = mask == 0
                background_val = np.mean(np_image[background_mask])

                for iX, cell_id in enumerate(cell_ids):
                    cell_mask = mask == cell_id
                    cell_val = np.mean(np_image[cell_mask])

                    cur_row_entries[iX][channel_name] = cell_val
                    cur_row_entries[iX][f"background {channel_name}"] = background_val
            row_entries += cur_row_entries

            kwargs = {"progress": str(int((iN + 1) / n_images * 100)) + "%",
                      "current_image": {"image_id": image_id}}
            if event_manager is None:
                self._call_update_listeners(**kwargs)
            else:
                event_manager.notify(ProgressEvent(percent=int((iN + 1) / n_images * 100),process=f"Readout Images: {iN+1}/{n_images} (Latest Image: {image_id})"))


        readout_path = os.path.join(self.directory, "readout.xlsx")
        df = pd.DataFrame(row_entries)
        df.to_excel(readout_path, index=False)
        kwargs = {}
        if event_manager is None:
            self._call_completion_listeners(readout=df, readout_path=readout_path, **kwargs)
        else:
            event_manager.notify(ProgressEvent(100,"Completed Readout"))



import numpy as np
from PyQt5.QtCore import QPointF

def mask_shifting(mask_data,deleted_mask_id:int,slice_id:int|None = None):
    """
    Shifts the mask when a mask got deleted to restore an order without gaps.

    Args:
        mask_data (np.array): the mask data.
        deleted_mask_id (int): the id of the deleted mask.
        slice_id (int): the id of the slice when the mask is 3d.

    Raises:
          ValueError: if the deleted_mask_id is smaller or equal to 0.
    """
    if deleted_mask_id < 1:
        raise ValueError("deleted_mask_id must be greater than 0")

    mask = mask_data["masks"]
    outline = mask_data["outlines"]

    if mask.ndim == 3:
        mask_slice = np.take(mask, int(slice_id if slice_id is not None else 0), axis=0)
        mask_slice[mask_slice>deleted_mask_id] -= 1
        mask[int(slice_id if slice_id is not None else 0), :, :] = mask_slice
    else:
        mask[mask > deleted_mask_id] -= 1

    if outline.ndim == 3:
        outline_slice = np.take(outline, int(slice_id if slice_id is not None else 0), axis=0)
        outline_slice[outline_slice>deleted_mask_id] -= 1
        outline[int(slice_id if slice_id is not None else 0), :, :] = outline_slice
    else:
        outline[outline>deleted_mask_id] -= 1

def search_free_id(mask,outline):
    """
    Search in a NumPy array of integers (e.g., [1,1,2,2,3,4,5,5,7,7]) for the first missing number (in this case, 6).
    If no gap is found, return the highest value + 1.
    """

    combined = np.concatenate((mask.ravel(),outline.ravel()))

    unique_vals = np.unique(combined)

    if unique_vals.size == 0:
        return 1

    diffs = np.diff(unique_vals)

    gap_index = np.where(diffs > 1)[0]

    if gap_index.size > 0:
        missed_value = unique_vals[gap_index[0]] + 1
    else:
        missed_value = unique_vals[-1] + 1

    return missed_value

def bresenham_line(start: QPointF, end: QPointF):
    """
    Calculates all pixel coordinates along a line from start to end using the Bresenham algorithm.
    """
    x0, y0 = int(round(start.x())), int(round(start.y()))
    x1, y1 = int(round(end.x())), int(round(end.y()))
    pixels = []

    dx = abs(x1 - x0)
    dy = abs(y1 - y0)
    sx = 1 if x0 < x1 else -1
    sy = 1 if y0 < y1 else -1
    err = dx - dy

    while True:
        pixels.append((x0, y0))
        if x0 == x1 and y0 == y1:
            break
        e2 = 2 * err
        if e2 > -dy:
            err -= dy
            x0 += sx
        if e2 < dx:
            err += dx
            y0 += sy

    return pixels

def trace_contour(binary_mask):
    """
    This method traces the contours of the inputted mask and returns the outline
    Attributes:
        binary_mask (np.array): the binary mask.

    Returns:
        contour: the outline of the cells in the mask

    """
    y_indices, x_indices = np.where(binary_mask)
    start_idx = np.lexsort((x_indices, y_indices))[0]
    start_x, start_y = x_indices[start_idx], y_indices[start_idx]

    directions = [(0, -1), (1, -1), (1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0), (-1, -1)]
    contour = []
    current = (start_x, start_y)
    prev = (start_x - 1, start_y)

    while True:
        contour.append(current)
        dir_idx = directions.index((prev[0] - current[0], prev[1] - current[1]))
        for i in range(8):
            next_dir = (dir_idx + 1 + i) % 8
            dx, dy = directions[next_dir]
            nx, ny = current[0] + dx, current[1] + dy
            if 0 <= ny < binary_mask.shape[0] and 0 <= nx < binary_mask.shape[1]:
                if binary_mask[ny, nx]:
                    prev = current
                    current = (nx, ny)
                    break
        if current == (start_x, start_y):
            break
    return contour

def fill_polygon_from_outline(contour, mask_shape):
    """
    Fills the given polygon in contour with the scanline technique.
    Attributes:
        contour (np.array): the contour
        mask_shape (np.array): the mask dimensions

    Returns:
        mask: the adapted mask with newly filled cells
    """
    mask = np.zeros(mask_shape, dtype=np.uint8)
    if not contour:
        return mask

    edges = []
    num_vertices = len(contour)
    for i in range(num_vertices):
        x0, y0 = contour[i]
        x1, y1 = contour[(i + 1) % num_vertices]
        if y0 == y1:
            continue  # Skip horizontal edges
        if y0 > y1:
            x0, y0, x1, y1 = x1, y1, x0, y0  # Ensure y0 < y1
        slope = (x1 - x0) / (y1 - y0) if (y1 - y0) != 0 else 0
        edges.append((y0, y1, x0, slope))

    if not edges:
        return mask

    global_edge_table = sorted(edges, key=lambda e: (e[0], e[2]))
    scan_line = min(e[0] for e in global_edge_table)
    active_edges = []

    while global_edge_table or active_edges:
        # Add edges starting at the current scan line
        while global_edge_table and global_edge_table[0][0] == scan_line:
            e = global_edge_table.pop(0)
            active_edges.append([e[1], e[2], e[3]])

        active_edges.sort(key=lambda e: e[1])

        # Fill between pairs of edges
        for i in range(0, len(active_edges), 2):
            if i + 1 >= len(active_edges):
                break
            e1 = active_edges[i]
            e2 = active_edges[i + 1]
            x_start = int(np.ceil(e1[1]))
            x_end = int(np.floor(e2[1]))
            for x in range(x_start, x_end + 1):
                if 0 <= x < mask_shape[1] and 0 <= scan_line < mask_shape[0]:
                    mask[scan_line, x] = 1

        scan_line += 1
        active_edges = [e for e in active_edges if e[0] > scan_line]
        for e in active_edges:
            e[1] += e[2]

    return mask

def find_border_pixels(mask, outline, cell_id,):
    """
    Finds edge pixels in a given mask matrix, considering the outline and only searching for pixels that match the specified cell_id.
    Edge pixels are those whose neighboring pixels have a different ID or are marked as edges in the outline.
    Attributes:
        mask: A 2D mask representing the ID of each pixel in the area.
        outline: A 2D mask marking the edge pixels (typically with a value of 100).
        cell_id: The ID of the cell whose edge pixels should be found.

    Returns:
        border_pixels: all neighbours of the cell
    """
    border_pixels = []

    rows, cols = mask.shape
    for y in range(rows):
        for x in range(cols):
            # ID of current pixel
            current_id = mask[y, x]

            if current_id != cell_id:
                continue

            # neighboring positions
            neighbors = [
                (y - 1, x), (y + 1, x),  # up, down
                (y, x - 1), (y, x + 1)  # left, right
            ]

            is_border_pixel = False

            for ny, nx in neighbors:
                if 0 <= ny < rows and 0 <= nx < cols:
                    if mask[ny, nx] != current_id and outline[ny, nx] != current_id:
                        is_border_pixel = True
                        break
            if is_border_pixel:
                border_pixels.append((y,x))

    return border_pixels




from abc import ABC, abstractmethod
from typing import List, Type, TypeVar, Generic
from cellsepi.backend.main_window.expert_mode.listener import *

class EventManager:
    """
    Manages the different Listeners and notify them if they associated event happens.
    """
    def __init__(self):
        self._listeners: dict[Type[Event], List[EventListener]] = {}

    def subscribe(self, listener: EventListener):
        """
        Adds a listener to the event manager.
        """
        event_type = listener.get_event_type()
        if event_type not in self._listeners:
            self._listeners[event_type] = []
        self._listeners[event_type].append(listener)

    def unsubscribe(self, listener: EventListener):
        """
        Removes a listener from the event manager.
        """
        event_type = listener.get_event_type()
        if event_type not in self._listeners or listener not in self._listeners[event_type]:
            raise ValueError("Listener is not subscribed")
        self._listeners[event_type].remove(listener)
        if len(self._listeners[event_type]) == 0:
            del self._listeners[event_type]

    def notify(self,event: Event):
        """
        Notify all listeners associated with the given event.
        """
        event_type = type(event)
        if event_type not in self._listeners:
            return
        for listener in self._listeners.get(event_type, []):
            listener.update(event)


import ast

import pytest

from src.cellsepi.backend.main_window.config_file import ConfigFile

from src.cellsepi.backend.main_window.config_file import create_default_config, DeletionForbidden


@pytest.fixture
def config():
    cfg = ConfigFile()
    cfg.clear_config()
    yield cfg
    cfg.restore_config()

def test_add_and_delete_profile(config):
    config.add_profile("test", "42", "n", "j", 2.0)
    default_config = create_default_config()
    assert config.config != default_config, "The config did not change"
    assert "test" in config.config["Profiles"], "The new profile was not added"
    assert "42" == config.config["Profiles"]["test"]["bf_channel"], "The bf_channel is not right"
    assert "n" == config.config["Profiles"]["test"]["mask_suffix"], "The mask_suffix is not right"
    assert "j" == config.config["Profiles"]["test"]["channel_prefix"], "The channel_prefix is not right"
    assert 2.0 == config.config["Profiles"]["test"]["diameter"], "The diameter is not right"
    config.delete_profile("test")
    assert config.config == default_config, "test is not deleted"

def test_if_config_is_empty(config):
    assert config.config == create_default_config(),"empty file did not construct right"

def test_if_config_is_not_existing(config):
    assert config.config == create_default_config(),"new file did not construct right"

def test_get_profile(config):
    assert config.get_profile("Lif") == create_default_config()["Profiles"]["Lif"], "Something get wrong by getting profile"

def test_update_profile(config):
    config.update_profile("Lif", "42")
    assert config.get_profile("Lif")["bf_channel"] == "42", "Values not changed after update"

def test_add_profile_name_fail(config):
    assert False == config.add_profile("Lif", 42, "n", "j", 2.0), "Adding should fail"

def test_rename_profile_name(config):
    assert True == config.rename_profile("Lif", "Lif2"), "Something went wrong with renaming"
    assert "Lif2" in config.config["Profiles"], "The profile name was not renamed"
    assert True == config.rename_profile("Lif2", "Lif2"), "Renaming should just return True because old and new are equal"
    assert False == config.rename_profile("Lif3","Lif2"), "Renaming should fail, because Lif3 not exists"
    assert False == config.rename_profile("Lif3", "Tif"), "Renaming should fail, because Tif is already taken"

def test_selected_profile_name(config):
    config.select_profile("Tif")
    assert config.get_selected_profile_name() == "Tif", "Selected profile was not changed"
    assert config.get_selected_profile() == create_default_config()["Profiles"]["Tif"], "Selected profile is wrong"

def test_delete_selected_profile_name(config):
    config.select_profile("Tif")
    config.delete_profile("Tif")
    assert config.get_selected_profile_name() == "Lif", "Selected profile was not changed after deleted"

def test_set_colors(config):
    config.set_mask_color((12,42,69))
    assert config.get_mask_color() == (12,42,69), "The mask_color is not right"
    config.set_outline_color((21,24,96))
    assert config.get_outline_color() == (21,24,96), "The outline_color is not right"

def test_states(config):
    config.set_auto_button(True)
    assert config.get_auto_button() == True, "The auto_button is not right"
    config.set_auto_button(False)
    assert config.get_auto_button() == False, "The auto_button is not right"
    config.set_lif_slider(True)
    assert config.get_lif_slider() == True, "The auto_button is not right"
    config.set_lif_slider(False)
    assert config.get_lif_slider() == False, "The auto_button is not right"

def test_attribute_getter(config):
    assert config.get_bf_channel() == create_default_config()["Profiles"]["Lif"]["bf_channel"], "bf_channel is wrong"
    assert config.get_mask_suffix() == create_default_config()["Profiles"]["Lif"]["mask_suffix"], "mask_suffix is wrong"
    assert config.get_channel_prefix() == create_default_config()["Profiles"]["Lif"]["channel_prefix"],"channel_prefix is wrong"
    assert config.get_diameter() == float(create_default_config()["Profiles"]["Lif"]["diameter"]),"diameter is wrong"
    assert config.get_mask_color() ==  ast.literal_eval(create_default_config()["Colors"]["mask"]),"mask_color is wrong"
    assert config.get_outline_color() ==  ast.literal_eval(create_default_config()["Colors"]["outline"]),"outline_color is wrong"

def test_idx_name(config):
    assert config.name_to_index("Lif") == 0, "Idx is wrong for Lif"
    assert config.index_to_name(0) == "Lif", "Lif is wrong for this IDX"

def test_invalid_profile(config):
        with pytest.raises(ValueError):
            config.add_profile("", "42", "", "", -10)
        with pytest.raises(ValueError):
            config.update_profile("Lif", "42", "", "", 2)
        with pytest.raises(ValueError):
            config.update_profile("Lif", "42", "a", "", 2)
        with pytest.raises(ValueError):
           config.update_profile("Lif", "42", "d", "a", -50)
        with pytest.raises(ValueError):
            config.rename_profile("","")
        with pytest.raises(ValueError):
            config.set_mask_color((22,12))
        with pytest.raises(ValueError):
            config.select_profile("")
        with pytest.raises(DeletionForbidden):
            config.delete_profile("Lif")
            config.delete_profile("Tif")



import threading

from cellsepi.backend.main_window.images import BatchImageReadout
from cellsepi.backend.main_window.cellsepi import CellSePi
from cellsepi.backend.main_window.notifier import Notifier
from cellsepi.frontend.main_window.gui_fluorescence import error_banner,fluorescence_button

class Fluorescence(Notifier):
    """
    class handles the readout of fluorescence values

    Attributes:
        csp= current CellSePi object
        gui= current gui object
    """
    def __init__(self,csp: CellSePi,gui):
        super().__init__()
        self.csp= csp
        self.gui=gui


    def readout_fluorescence(self):
        """
        starts the readout of fluorescence and creates an Excel list if possible

        """

        if self.check_readout_possible():
            self._call_start_listeners(True)
            def on_update(progress,current_image):
                self._call_update_listeners(progress, current_image)

            def completed_readout(readout, readout_path):
                self.csp.readout= readout
                self.csp.readout_path=readout_path
                self.csp.readout_running=False
                if readout_path is not None:
                    self.gui.open_button.visible=True
                self.gui.page.run_task(self.gui.directory.check_masks)
                fluorescence_button.disabled = False
                if self.csp.model_path is not None:
                    self.gui.start_button.disabled =False
                self._call_completion_listeners()


            self.csp.readout_running=True
            fluorescence_button.disabled=True

            brightfield_channel = self.csp.config.get_bf_channel()
            prefix = self.csp.current_channel_prefix
            working_directory = self.csp.working_directory

            #creates the readout image and fills the mask_path
            batch_image_readout = BatchImageReadout(image_paths=self.csp.image_paths,
                                                          mask_paths=self.csp.mask_paths,
                                                          segmentation_channel=brightfield_channel,
                                                          channel_prefix=prefix,
                                                          directory=working_directory)
            batch_image_readout.add_update_listener(listener=on_update)
            batch_image_readout.add_completion_listener(listener=completed_readout)

            target = batch_image_readout.run
            self.csp.readout_thread = threading.Thread(target=target)

            self._call_start_listeners(False)
            self.csp.readout_thread.start()


        else:
            self._call_start_listeners(False)
            return


    def check_readout_possible(self):
        """
         -error handling. All error that can possibly occur
         -creates a banner visible in the GUI if an error is important
        """

        if self.csp.readout_running:
            return False,
        if self.csp.readout_thread is not None and self.csp.readout_thread.is_alive():
            return False
        if self.csp.image_paths is None or len(self.csp.image_paths) ==0 :
            error_banner(self.gui, "No image to process")
            return False

        return True





import textwrap
from typing import Type
import flet as ft
from cellsepi.backend.main_window.expert_mode.listener import EventListener, OnPipelineChangeEvent, Event, \
    ModuleExecutedEvent, ProgressEvent, ErrorEvent, ModuleStartedEvent, DragAndDropEvent, PipelinePauseEvent, \
    PipelineCancelEvent, PipelineErrorEvent

class PipelineChangeListener(EventListener):
    """
    Gets notified when a OnPipelineChangeEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self,builder):
        self.event_type = OnPipelineChangeEvent
        self.builder = builder
    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        if self.builder.pipeline_gui.loading:
            return
        if event.change_type == "user_attr_change":
            if len(self.builder.pipeline_gui.pipeline.modules)-len(self.builder.pipeline_gui.show_room_modules) > 0:
                if self.builder.pipeline_gui.pipeline_directory != "":
                    if self.builder.pipeline_storage.check_saved():
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}"
                        self.builder.save_button.icon_color = ft.Colors.WHITE24
                        self.builder.save_button.disabled = True
                        self.builder.page.update()
                    else:
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}*"
                        self.builder.save_button.icon_color = ft.Colors.WHITE60
                        self.builder.save_button.disabled = False
                        self.builder.page.update()
                else:
                    if self.builder.pipeline_storage.check_saved():
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}"
                        self.builder.page.update()
                    else:
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}*"
                        self.builder.page.update()
        else:
            if len(self.builder.pipeline_gui.pipeline.modules)-len(self.builder.pipeline_gui.show_room_modules) > 0:
                self.builder.help_text.opacity = 0
                self.builder.help_text.update()
                self.builder.save_as_button.icon_color = ft.Colors.WHITE60
                self.builder.save_as_button.disabled = False
                self.builder.save_as_button.update()
                if self.builder.pipeline_gui.pipeline_directory != "":
                    if self.builder.pipeline_storage.check_saved():
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}"
                        self.builder.save_button.icon_color = ft.Colors.WHITE24
                        self.builder.save_button.disabled = True
                        self.builder.page.update()
                    else:
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}*"
                        self.builder.save_button.icon_color = ft.Colors.WHITE60
                        self.builder.save_button.disabled = False
                        self.builder.page.update()
                else:
                    if self.builder.pipeline_storage.check_saved():
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}"
                        self.builder.page.update()
                    else:
                        self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}*"
                        self.builder.page.update()


                if not self.builder.pipeline_gui.pipeline.running:
                    self.builder.start_button.disabled = False
                    self.builder.start_button.update()
            else:
                self.builder.help_text.opacity = 1
                self.builder.help_text.update()
                self.builder.save_as_button.icon_color = ft.Colors.WHITE24
                self.builder.save_as_button.disabled = True
                self.builder.save_button.icon_color = ft.Colors.WHITE24
                self.builder.save_button.disabled = True
                self.builder.start_button.disabled = True
                self.builder.page.title = f"CellSePi - {self.builder.pipeline_gui.pipeline_name}"
                self.builder.page.update()
        if self.builder.pipeline_gui.pipeline.running:
            self.builder.update_modules_executed()
        else:
            self.builder.update_modules_executed(reset=True)

class DragAndDropListener(EventListener):
    """
    Gets notified when a DragAndDropEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self,builder):
        self.event_type = DragAndDropEvent
        self.builder = builder
    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        if len(self.builder.pipeline_gui.modules) == 0:
            if event.drag:
                self.builder.help_text.opacity = 0.60
                self.builder.help_text.update()
            else:
                self.builder.help_text.opacity = 1
                self.builder.help_text.update()

class ModuleExecutedListener(EventListener):
    """
    Gets notified when a ModuleExecutedEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self,builder):
        self.event_type = ModuleExecutedEvent
        self.builder = builder
    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        self.builder.pipeline_gui.modules[event.module_id].enable_tools()
        if self.builder.pipeline_gui.source_module != "":
            self.builder.pipeline_gui.check_for_valid(event.module_id)
        self.builder.pipeline_gui.lines_gui.update_delete_buttons(self.builder.pipeline_gui.modules[event.module_id])
        self.builder.pipeline_gui.check_all_deletable()
        self.builder.pipeline_gui.modules[event.module_id].executing_button.visible = False
        self.builder.pipeline_gui.modules[event.module_id].waiting_button.visible = False
        self.builder.pipeline_gui.modules[event.module_id].executing_button.update()
        self.builder.pipeline_gui.modules[event.module_id].waiting_button.update()
        self.builder.update_modules_executed()


class ModuleStartedListener(EventListener):
    """
    Gets notified when a ModuleStartedEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self,builder):
        self.event_type = ModuleStartedEvent
        self.builder = builder

    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        self.builder.pipeline_gui.modules[event.module_id].set_running()
        self.builder.category_icon.color = self.builder.pipeline_gui.modules[event.module_id].module.gui_config().category.value
        self.builder.category_icon.update()
        self.builder.running_module.value = self.builder.pipeline_gui.modules[event.module_id].module.gui_config().name
        self.builder.running_module.update()

class PipelinePauseListener(EventListener):
    """
    Gets notified when a PipelinePauseEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self,builder):
        self.event_type = PipelinePauseEvent
        self.builder = builder
        self.last_info_text = ""
        self.last_info_spans = []
        self.last_running_module = ""

    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        if event.resume:
            self.builder.pipeline_gui.modules[event.module_id].disable_pause()
            self.builder.pipeline_gui.modules[event.module_id].paused_button.visible = False
            self.builder.pipeline_gui.modules[event.module_id].paused_button.update()
            self.builder.pipeline_gui.modules[event.module_id].executing_button.visible = True
            self.builder.pipeline_gui.modules[event.module_id].executing_button.update()
            self.builder.info_text.value = self.last_info_text
            self.builder.info_text.spans = self.last_info_spans
            self.builder.info_text.update()
            self.builder.running_module.value = self.last_running_module
            self.builder.running_module.update()
            self.builder.cancel_button.visible = True
            self.builder.cancel_button.update()
            self.builder.resume_button.visible = False
            self.builder.resume_button.update()
            return

        self.builder.pipeline_gui.modules[event.module_id].enable_pause()
        self.builder.pipeline_gui.modules[event.module_id].paused_button.visible = True
        self.builder.pipeline_gui.modules[event.module_id].paused_button.update()
        self.builder.pipeline_gui.modules[event.module_id].executing_button.visible = False
        self.builder.pipeline_gui.modules[event.module_id].executing_button.update()
        self.last_running_module = self.builder.running_module.value
        self.builder.running_module.value = f"Paused: {self.builder.running_module.value}"
        self.builder.running_module.update()
        self.last_info_text = self.builder.info_text.value
        self.last_info_spans = self.builder.info_text.spans.copy()
        self.builder.info_text.value = "Pipeline paused: press resume button to carry on"
        self.builder.info_text.spans = []
        self.builder.info_text.update()
        self.builder.resume_button.visible = True
        self.builder.resume_button.update()
        self.builder.cancel_button.visible = False
        self.builder.cancel_button.update()

class ModuleProgressListener(EventListener):
    """
    Gets notified when a ProgressEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self, builder):
        self.event_type = ProgressEvent
        self.builder = builder

    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        if self.builder.pipeline_gui.pipeline._cancel_event.is_set():
            self.builder.info_text.value = ""
            self.builder.info_text.spans = [
                ft.TextSpan("Canceling: ", style=ft.TextStyle(weight=ft.FontWeight.BOLD, color=ft.Colors.RED)),
                ft.TextSpan(event.process, style=ft.TextStyle(color=ft.Colors.WHITE60)), ]
            self.builder.info_text.update()
        else:
            self.builder.info_text.value = event.process
            self.builder.info_text.spans = []
            self.builder.info_text.update()
        self.builder.progress_bar_module.value = event.percent / 100
        self.builder.progress_bar_module.update()
        self.builder.progress_bar_module_text.value = f"{event.percent}%"
        self.builder.page.update()

class ModuleErrorListener(EventListener):
    """
    Gets notified when a ErrorEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self, builder):
        self.event_type = ErrorEvent
        self.builder = builder

    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
            self.builder.cancel_button.visible = False
            self.builder.cancel_button.disabled = False
            self.builder.cancel_button.color = ft.Colors.RED
            self.builder.cancel_button.update()
            self.builder.pipeline_gui.modules[self.builder.pipeline_gui.pipeline.executing].error_stack.visible = True
            self.builder.pipeline_gui.modules[self.builder.pipeline_gui.pipeline.executing].error_stack.update()
            wrapped_text = "\n".join(textwrap.wrap(event.error_msg, width=30))
            self.builder.pipeline_gui.modules[self.builder.pipeline_gui.pipeline.executing].error_icon.tooltip = f"An error occurred while executing!\nError: {wrapped_text}"
            self.builder.pipeline_gui.modules[self.builder.pipeline_gui.pipeline.executing].error_icon.update()
            self.builder.pipeline_gui.modules[
                self.builder.pipeline_gui.pipeline.executing].module_container.border = ft.border.all(4,
                                                                                                      ft.Colors.RED)
            self.builder.pipeline_gui.modules[self.builder.pipeline_gui.pipeline.executing].module_container.update()
            self.builder.pipeline_gui.enables_all_stuck_in_running()
            self.builder.info_text.value = ""
            self.builder.info_text.spans = [
            ft.TextSpan("Error: ", style=ft.TextStyle(weight=ft.FontWeight.BOLD, color=ft.Colors.RED)),
            ft.TextSpan(event.error_msg, style=ft.TextStyle(color=ft.Colors.WHITE60)),]
            self.builder.info_text.update()
            self.builder.category_icon.color = ft.Colors.RED
            self.builder.category_icon.update()
            self.builder.progress_bar_module_text.value = f"{0}%"
            self.builder.progress_bar_module.value = 0
            self.builder.page.update()


class PipelineCancelListener(EventListener):
    """
    Gets notified when a PipelineCancelEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self, builder):
        self.event_type = PipelineCancelEvent
        self.builder = builder

    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        self.builder.info_text.spans = []
        self.builder.info_text.value = "Idle, waiting for start."
        self.builder.info_text.update()
        self.builder.pipeline_gui.enables_all_stuck_in_running()
        self.builder.running_module.value = "Module"
        self.builder.running_module.update()
        self.builder.category_icon.color = ft.Colors.GREEN
        self.builder.category_icon.update()
        self.builder.resume_button.visible = False
        self.builder.resume_button.update()
        self.builder.cancel_button.visible = True
        self.builder.cancel_button.disabled = False
        self.builder.cancel_button.color = ft.Colors.RED
        self.builder.cancel_button.update()
        self.builder.progress_bar_module_text.value = f"{0}%"
        self.builder.progress_bar_module.value = 0
        self.builder.page.update()

class PipelineErrorListener(EventListener):
    """
    Gets notified when a PipelineErrorEvent is happening and updates all GUI elements related to the event.
    """
    def __init__(self, builder):
        self.event_type = PipelineErrorEvent
        self.builder = builder

    def get_event_type(self) -> Type[Event]:
        return self.event_type

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    def _update(self, event: Event) -> None:
        self.builder.info_text.value = ""
        self.builder.info_text.spans = [
            ft.TextSpan("Error: ", style=ft.TextStyle(weight=ft.FontWeight.BOLD, color=ft.Colors.RED)),
            ft.TextSpan(event.error_msg, style=ft.TextStyle(color=ft.Colors.WHITE60)), ]
        self.builder.info_text.update()
        self.builder.pipeline_gui.enables_all_stuck_in_running()
        self.builder.running_module.value = "Pipeline"
        self.builder.running_module.update()
        self.builder.category_icon.color = ft.Colors.RED
        self.builder.category_icon.update()
        self.builder.progress_bar_module_text.value = f"{0}%"
        self.builder.progress_bar_module.value = 0
        self.builder.page.update()



import os
import platform
import re
import subprocess
import sys
import flet as ft

from cellsepi.backend.main_window.avg_diameter import AverageDiameter
from cellsepi.backend.main_window.fluorescence import Fluorescence
from cellsepi.frontend.main_window.gui_fluorescence import fluorescence_button
from cellsepi.backend.main_window.segmentation import Segmentation
from cellsepi.frontend.main_window.gui_mask import handle_mask_update, reset_mask


class GUISegmentation():
    """
    This class handles the segmentation, which can be controlled by the start, pause, resume and cancel buttons.
    """
    def __init__(self, gui):
        self.gui = gui
        self.segmentation = Segmentation(self, gui)
        self.fluorescence = Fluorescence(gui.csp, gui)
        self.segmentation_cancelling = False
        self.segmentation_pausing = False
        self.segmentation_resuming = False
        self.segmentation_currently_paused = False

    def create_segmentation_card(self):
        """
        This method creates a segmentation card for the GUI, which contains the progress bar and several buttons for
         controlling the run of the segmentation.

        Returns:
            segmentation_card (ft.Card): the card containing all the elements needed to run the segmentation
        """
        # creating all the necessary buttons and their initial properties
        start_button = ft.ElevatedButton( # button to start the segmentation calculation
            text="Start",
            icon=ft.Icons.PLAY_CIRCLE,
            tooltip="Start the segmentation",
            disabled=True,
            on_click=None
        )
        pause_button = ft.ElevatedButton( # button to pause the segmentation calculation while it is running
            text="Pause",
            icon=ft.Icons.PAUSE_CIRCLE,
            visible=False,
            on_click=None,
        )
        cancel_button = ft.ElevatedButton( # button to completely cancel the currently running segmentation calculation
            text="Cancel",
            icon=ft.Icons.CANCEL,
            visible=False,
            on_click=None,
            color=ft.Colors.RED,
            icon_color=ft.Colors.RED,
        )
        resume_button = ft.ElevatedButton( # button to resume the segmentation calculation after it has been paused
            text="Resume",
            icon=ft.Icons.PLAY_CIRCLE,
            visible=False,
            on_click=None,
        )

        # button to start the fluorescence readout
        fl_button = fluorescence_button

        # progress bar, which is updated throughout the segmentation calculation and fluorescence readout
        progress_bar = ft.ProgressBar(value=0, width=180)
        progress_bar_text = ft.Text("Waiting for Input")

        def open_readout(e):
            file_path = self.gui.csp.readout_path
            if os.name == "nt":  # Check if Windows
                os.startfile(file_path)
            elif os.name == "posix":  # Check if Mac or Linux
                subprocess.run(["open", file_path] if sys.platform == "darwin" else ["xdg-open", file_path])

        open_button = ft.IconButton(
            icon=ft.Icons.OPEN_IN_NEW_ROUNDED,
            tooltip = "Open fluorescence file",
            on_click=open_readout,
            visible= False
        )

        # the following methods are called when clicking on the corresponding button
        def pick_model_result(e: ft.FilePickerResultEvent):
            """
            The result of the file selection is handled.

            Arguments:
                e (ft.FilePickerResultEvent): the result of the file picker event, i.e. the chosen file
            """
            if e.files is None:
                #case: no model selected
                pass
            elif e.files[0].path is not None:
                if self.gui.ready_to_start:
                    progress_bar_text.value = "Ready to Start"
                    start_button.disabled = False
                model_text.value = e.files[0].name
                model_text.color = None
                self.gui.csp.model_path = e.files[0].path
                self.gui.page.update()

        pick_model_dialog = ft.FilePicker(on_result=pick_model_result)
        self.gui.page.overlay.extend([pick_model_dialog])

        def start_segmentation(e): # called when the start button is clicked
            """
            The start of the segmentation is initialized.
            This includes error handling for the case where something other than a model is chosen.
            """
            # visibility of buttons before start of segmentation (needed in case of error)
            state_fl_button = fl_button.visible
            state_open_button = self.gui.open_button.visible
            try:
                start_button.visible = False
                pause_button.visible = True
                cancel_button.visible = True
                model_title.disabled = True
                model_chooser.disabled = True
                fl_button.visible = False
                cancel_button.color = ft.Colors.RED
                cancel_button.icon_color = ft.Colors.RED
                self.gui.open_button.visible = False
                self.gui.training_environment.disable_switch_environment()
                self.gui.directory.disable_path_choosing()
                self.gui.page.update()
                self.segmentation.run() # this will throw an error if something other than a model was chosen
            except:
                self.gui.page.open(ft.SnackBar(ft.Text("You have selected an incompatible file for the segmentation model.")))
                self.gui.training_environment.enable_switch_environment()
                start_button.visible = True
                start_button.disabled = True
                pause_button.visible = False
                cancel_button.visible = False
                model_title.disabled = False
                model_text.color = ft.Colors.RED
                model_chooser.disabled = False
                fl_button.visible = state_fl_button
                self.gui.open_button.visible = state_open_button
                self.gui.directory.enable_path_choosing()
                self.gui.csp.segmentation_running = False
                progress_bar_text.value = "Select new Model"
                self.gui.csp.model_path = None
                self.gui.page.update()


        def cancel_segmentation(): # called when the cancel button is clicked
            """
            The running segmentation is cancelled and everything returns to the start state.
            The masks calculated so far are deleted and the previously calculated masks are restored.
            """
            cancel_button.visible = False
            self.segmentation_cancelling = True
            self.segmentation.to_be_cancelled()
            self.gui.diameter_display.opacity = 0.5
            if self.segmentation_currently_paused:
                resume_button.visible = False
                extracted_percentage = re.search(r'\d+', progress_bar_text.value)
                progress_bar_text.value = "Cancelling: " + extracted_percentage.group(0) + " %"
                self.segmentation_currently_paused = False
                self.gui.page.update()
                self.segmentation.run()
            else:
                pause_button.visible = False
                progress_bar_text.value = "Cancelling: " + progress_bar_text.value
            self.gui.page.update()


        def pause_segmentation(e): # called when the pause button is clicked
            """
            The running segmentation is paused (the progress of the segmentation so far is stored).
            """
            pause_button.visible = False
            resume_button.visible = True
            resume_button.disabled = True
            cancel_button.disabled = True
            progress_bar_text.value = "Pausing: " + progress_bar_text.value
            cancel_button.color = None
            cancel_button.icon_color = None
            self.gui.page.update()
            self.segmentation_pausing = True
            self.segmentation.to_be_paused()

        def resume_segmentation(e): # called when the resume button is clicked
            """
            The segmentation is resumed again from the previously paused state.
            """
            self.segmentation_currently_paused = False
            resume_button.visible = False
            pause_button.visible = True
            pause_button.disabled = True
            cancel_button.disabled = True
            extracted_percentage = re.search(r'\d+', progress_bar_text.value)
            progress_bar_text.value =  extracted_percentage.group(0) + " %" # remove "paused at " from string
            cancel_button.color = None
            cancel_button.icon_color = None
            self.gui.page.update()
            self.segmentation.to_be_resumed()
            self.segmentation_resuming = True
            self.segmentation.run()


        # define behavior of buttons when they are clicked
        start_button.on_click = start_segmentation
        cancel_button.on_click = lambda e: cancel_segmentation()
        pause_button.on_click = pause_segmentation
        resume_button.on_click = resume_segmentation

        def finished_segmentation():
            """
            This method updates the segmentation card when the segmentation is finished.
            After this a new segmentation can be started again.
            """
            progress_bar_text.value = "Finished"
            pause_button.visible = False
            cancel_button.visible = False
            fl_button.visible = True
            fl_button.disabled = False
            start_button.visible = True
            start_button.disabled = False
            model_title.disabled = False
            model_chooser.disabled = False
            self.gui.diameter_text.value = self.gui.average_diameter.get_avg_diameter()
            self.gui.training_environment.enable_switch_environment()
            self.gui.directory.enable_path_choosing()
            self.gui.csp.segmentation_running = False
            self.gui.page.update()

        def cancelled_segmentation():
            """
            This method updates the segmentation card when the cancellation is finished.
            """
            progress_bar_text.value = "Ready to Start"
            progress_bar.value = 0
            if not platform.system() == "Linux":
                self.gui.page.window.progress_bar = -1
            start_button.visible = True
            self.gui.page.run_task(self.gui.directory.check_masks)
            if self.gui.csp.readout_path is not None:
                self.gui.open_button.visible = True
            self.segmentation_cancelling = False
            self.gui.training_environment.enable_switch_environment()
            self.gui.directory.enable_path_choosing()
            self.gui.csp.segmentation_running = False
            model_title.disabled = False
            model_chooser.disabled = False
            self.gui.diameter_display.opacity = 1
            self.gui.diameter_text.value = self.gui.average_diameter.get_avg_diameter()
            for image_id in self.gui.csp.image_paths:
                self.gui.directory.update_mask_check(image_id)
            self.gui.page.update()
            if self.gui.cancel_event is not None:
                self.gui.cancel_event.set()

        def paused_segmentation():
            """
            This method updates the segmentation card when the pausing is finished.
            """
            resume_button.disabled = False
            cancel_button.disabled = False
            extracted_percentage = re.search(r'\d+', progress_bar_text.value)
            progress_bar_text.value = "Paused at: " + extracted_percentage.group(0) + " %"
            cancel_button.color = ft.Colors.RED
            cancel_button.icon_color = ft.Colors.RED
            self.gui.page.update()
            self.segmentation_pausing = False
            self.segmentation_currently_paused = True

        def resumed_segmentation():
            """
            This method updates the segmentation card when the resuming has successfully started.
            """
            pause_button.disabled = False
            cancel_button.disabled = False
            cancel_button.color = ft.Colors.RED
            cancel_button.icon_color = ft.Colors.RED
            self.gui.page.update()
            self.segmentation_resuming = False

        def update_progress_bar(progress,current_image):
            """
            This method updates the progress bar at any point before, during and after the segmentation and fluorescence process.

            Arguments:
                progress (int): the current progress
                current_image (dict): the current image number
            """
            if self.segmentation_pausing:
                progress_bar_text.value = "Pausing: " + str(progress)
            elif self.segmentation_cancelling:
                progress_bar_text.value = "Cancelling: " + str(progress)
                self.gui.diameter_display.opacity = 0.5
            else:
                progress_bar_text.value = progress
            extracted_num = re.search(r'\d+', progress)
            if extracted_num is not None:
                progress_bar.value = int(extracted_num.group())/100
                if not platform.system() == "Linux":
                    self.gui.page.window.progress_bar = progress_bar.value

            if current_image is not None:
                if current_image["image_id"] == self.gui.csp.window_image_id:
                    if self.segmentation.batch_image_segmentation.segmentation_channel == self.gui.csp.window_bf_channel:
                        self.gui.queue.put("refresh_mask")
                if current_image["image_id"] == self.gui.csp.image_id and self.segmentation.batch_image_segmentation.segmentation_channel == self.gui.csp.config.get_bf_channel():
                    handle_mask_update(self.gui)
                else:
                    reset_mask(self.gui, current_image["image_id"], self.segmentation.batch_image_segmentation.segmentation_channel)
            self.gui.page.update()

        # listeners for getting different information from the state of the segmentation process
        self.segmentation.add_update_listener(listener=update_progress_bar)
        self.segmentation.add_completion_listener(listener=finished_segmentation)
        self.segmentation.add_cancel_listener(listener=cancelled_segmentation)
        self.segmentation.add_pause_listener(listener=paused_segmentation)
        self.segmentation.add_resume_listener(listener=resumed_segmentation)

        def fluorescence_readout(e):
            """
            Updates the segmentation card and starts readout of the fluorescence data when the fluorescence button is clicked.
            """
            self.fluorescence.readout_fluorescence()
            fl_button.disabled = True
            start_button.disabled = True
            self.gui.open_button.visible = False
            progress_bar_text.value = "Reading fluorescence"
            self.gui.directory.disable_path_choosing()
            model_title.disabled = True
            model_chooser.disabled = True
            self.gui.page.update()

        def start_fl(e):
            progress_bar.value = 0
            if not platform.system() == "Linux":
                self.gui.page.window.progress_bar = 0
            progress_bar_text.value = "0 %"
            self.gui.page.update()

        def complete_fl():
            progress_bar.value = 0
            if not platform.system() == "Linux":
                self.gui.page.window.progress_bar = -1
            if self.gui.csp.model_path is not None:
                progress_bar_text.value = "Ready to start"
            else:
                progress_bar_text.value = "Waiting for Input"
            self.gui.directory.enable_path_choosing()
            model_title.disabled = False
            model_chooser.disabled = False
            self.gui.page.update()
            if self.gui.readout_event is not None:
                self.gui.readout_event.set()


        fl_button.on_click = fluorescence_readout
        self.fluorescence.add_start_listener(listener=start_fl)
        self.fluorescence.add_update_listener(listener=update_progress_bar)
        self.fluorescence.add_completion_listener(listener=complete_fl)

        pick_model_row = ft.Row(
            [
                ft.Container(content=ft.Row([progress_bar, progress_bar_text])),
                ft.Container(content=ft.Row([start_button, pause_button, resume_button, cancel_button, fl_button, open_button]))
            ], alignment=ft.MainAxisAlignment.SPACE_BETWEEN
        )
        model_text = ft.Text("Choose Model")
        model_title = ft.ListTile(
                                        leading=ft.Icon(name=ft.Icons.HUB_OUTLINED),
                                        title= model_text,
                                    )

        segmentation_container = ft.Container(
                            content=ft.Column(
                                [model_title,
                                 pick_model_row,
                                 ]
                            )
                        )

        project_root =os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        model_directory = os.path.join(project_root, "models")

        model_chooser = ft.Container(
                            content=ft.IconButton(
                                icon=ft.Icons.UPLOAD_FILE,
                                tooltip="Choose model",
                                on_click=lambda _: pick_model_dialog.pick_files(allow_multiple=False, initial_directory=model_directory),
                            ), alignment=ft.alignment.bottom_right,
                        )

        segmentation_card = ft.Card(
            content=ft.Container(
                content=ft.Stack(
                    [   segmentation_container,
                        model_chooser
                    ]
                ),
                padding=10
            ),
        )
        return segmentation_card,start_button,open_button,progress_bar,progress_bar_text,cancel_segmentation


import asyncio

import flet as ft
import tifffile

from cellsepi.frontend.main_window.gui_mask import handle_image_switch_mask_on

def update_main_image(img_id,channel_id,gui,on_click = True):
    """
    Method that handles what happens when the image is clicked or the main image need an update.
    """
    if on_click:
        if gui.csp.image_id is not None and gui.csp.image_id in gui.directory.selected_images_visualise:
            if gui.csp.channel_id is not None and gui.csp.channel_id in gui.directory.selected_images_visualise[gui.csp.image_id]:
                gui.directory.selected_images_visualise[gui.csp.image_id][gui.csp.channel_id].visible = False
                gui.directory.selected_images_visualise[gui.csp.image_id][gui.csp.channel_id].update()
    gui.csp.image_id = img_id
    gui.csp.channel_id = channel_id
    gui.directory.selected_images_visualise[img_id][channel_id].visible = True
    gui.directory.selected_images_visualise[img_id][channel_id].update()
    handle_image_switch_mask_on(gui)
    if on_click:
        gui.contrast_slider.value = 1.0
        gui.brightness_slider.value = 1.0
    gui.contrast_slider.update()
    gui.brightness_slider.update()
    if img_id is not None:
        if tifffile.imread(gui.csp.image_paths[img_id][channel_id]).ndim == 3: #check if 3d, because z_max projection is not editable in the drawing tool only 2.5 sliced 3d images.
            gui.drawing_button.disabled = True
        else:
            gui.drawing_button.disabled = False
    else:
        gui.drawing_button.disabled = True
    gui.drawing_button.update()
    if not gui.auto_image_tuning.active:
        gui.contrast_slider.disabled = False
        gui.brightness_slider.disabled = False
        gui.page.update()
        asyncio.run(
            gui.image_tuning.update_brightness_and_contrast_async(on_click=on_click, linux_or_3d=gui.csp.linux_or_3d))
    else:
        gui.auto_image_tuning.update_main_image_auto(image_path = gui.csp.image_paths[gui.csp.image_id][gui.csp.channel_id])



class Canvas:
    """
    Creates the main_card with the main_image and mask.
    """
    def __init__(self):
        self.container_mask=ft.Container(ft.Image(src=r"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA\AAAFCAIAAAFe0wxPAAAAAElFTkSuQmCC",fit=ft.ImageFit.SCALE_DOWN,),visible=False,alignment=ft.alignment.center)

        self.main_image = ft.Container(ft.Image(src=r"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA\AAAFCAIAAAFe0wxPAAAAAElFTkSuQmCC",
                                    fit=ft.ImageFit.SCALE_DOWN),alignment=ft.alignment.center)

        self.canvas_card = self.create_canvas_card()
    def create_canvas_card(self):
        return ft.Card(
            content=ft.Stack([self.main_image, self.container_mask]),
            expand=True
        )




import math
from collections import deque
from typing import List

import flet as ft

from cellsepi.backend.main_window.expert_mode.listener import OnPipelineChangeEvent
from cellsepi.backend.main_window.expert_mode.pipe import Pipe
from cellsepi.backend.main_window.expert_mode.pipeline import Pipeline
from cellsepi.frontend.main_window.expert_mode.expert_constants import *
from cellsepi.frontend.main_window.expert_mode.gui_lines import LinesGUI
from cellsepi.frontend.main_window.expert_mode.gui_module import ModuleGUI


class PipelineGUI(ft.Stack):
    def __init__(self,page:ft.Page):
        super().__init__()
        self.controls = []
        self.pipeline = Pipeline()
        self.module_count = 0 #without show_room modules
        self.loading = False
        self.interactive_view = None
        self.pipeline_name = "NewPipeline"
        self.pipeline_directory = ""
        self.pipeline_dict = {} #last saved pipeline dict
        self.page = page
        self.modules = {} #without show_room modules, identiefierer is the module_id
        self.show_room_size = len(ModuleType) if len(ModuleType) < SHOWROOM_MODULE_COUNT else SHOWROOM_MODULE_COUNT
        self.show_room_modules = [] #saves all modules within the show_room
        self.width = BUILDER_WIDTH
        self.height = BUILDER_HEIGHT
        self.source_module: str = ""
        self.show_ports:bool = False
        self.show_delete_button:bool = False
        self.transmitting_ports: List[str] = []
        self.lines_gui = LinesGUI(self)
        self.controls.append(self.lines_gui)
        self.show_room_container = None
        self.show_room_page_number: int = 0
        self.show_room_max_page_number: int = 0
        self.page_stack = None
        self.delete_stack = ft.Stack()
        self.controls.append(self.delete_stack)
        self.expand = True

    def reset(self):
        """
        Resets the PipelineGUI, so a other pipeline can be loaded into.
        """
        self.loading = True
        for module in list(self.modules.values()):
            module.remove_module()
        self.pipeline.run_order = deque()
        self.pipeline.executing = ""
        self.pipeline.running = False
        self.update()
        self.loading = False

    def load_pipeline(self):
        """
        Loads a pipeline from a dict to PipelineGUI.
        """
        self.loading = True
        for module_dict in self.pipeline_dict["modules"]:
            type_map = {mt.value.gui_config().name: mt for mt in ModuleType}
            if module_dict["module_name"] not in type_map:
                self.loading = False
                self.pipeline.event_manager.notify(OnPipelineChangeEvent(f"Pipeline {self.pipeline_name} loaded."))
                raise ValueError(f"Module {module_dict['module_name']} not supported!")
            self.add_module(module_type=type_map[module_dict["module_name"]], x=module_dict["position"]["x"], y=module_dict["position"]["y"], module_id=module_dict["module_id"],module_dict=module_dict)

        for pipe in self.pipeline_dict["pipes"]:
            source = pipe["source"]
            target = pipe["target"]
            ports= pipe["ports"]
            self.add_connection(self.modules[source],self.modules[target],ports)

        offset_x = self.pipeline_dict["view"]["offset_x"]
        offset_y = self.pipeline_dict["view"]["offset_y"]
        scale = self.pipeline_dict["view"]["scale"]
        self.interactive_view.set_transformation_data(offset_x=offset_x, offset_y=offset_y, scale=scale)

        self.page.open(
            ft.SnackBar(ft.Text(f"Pipeline successfully loaded.", color=ft.Colors.WHITE), bgcolor=ft.Colors.GREEN))
        self.page.update()
        self.loading = False
        self.pipeline.event_manager.notify(OnPipelineChangeEvent(f"Pipeline {self.pipeline_name} loaded."))

    def check_all_deletable(self):
        """
        Checks for all modules if their delete_button can be maked visible.
        This is the case if the pipeline is no longer executing or all direct connection of a module are executed.
        """
        for module in self.modules.values():
            if self.check_deletable(module):
                module.delete_button.visible = True
                module.delete_button.update()
            else:
                module.delete_button.visible = False
                module.delete_button.update()

    def check_deletable(self,module:ModuleGUI):
        """
        Checks if the pipeline is no longer executing or all direct connection of the given module are executed.
        """
        if  module.module_id in self.pipeline.run_order or module.module_id == self.pipeline.executing:
            return False

        return all(not(pipe.target_module.module_id in self.pipeline.run_order or pipe.target_module.module_id == self.pipeline.executing) for pipe in self.pipeline.pipes_out[module.module_id])

    def add_connection(self,source_module_gui,target_module_gui,ports: List[str]):
        """
        Adds a connection to the pipeline.
        """
        ports_copy = list(ports)
        self.pipeline.add_connection(pipe=Pipe(source_module_gui.module, target_module_gui.module, ports_copy))
        self.lines_gui.update_line(source_module_gui, target_module_gui,ports)
        self.lines_gui.update_gui()
        self.update_all_port_icons()

    def expand_connection(self,pipe:Pipe,ports:List[str]):
        """
        Expands a connection, so its connection also tranferes the given ports.
        """
        ports_copy = list(ports)
        self.pipeline.expand_connection(pipe,ports_copy)
        self.lines_gui.update_line(self.modules[pipe.source_module.module_id], self.modules[pipe.target_module.module_id], pipe.ports)
        self.lines_gui.update_gui()
        self.update_all_port_icons()

    def remove_connection(self,source_module_gui,target_module_gui):
        """
        Removes a connection from the pipeline.
        """
        self.pipeline.remove_connection(source_module_gui.module_id,target_module_gui.module_id)
        self.lines_gui.remove_line(source_module_gui, target_module_gui)
        self.update_all_port_icons()
        self.check_for_valid_all_modules()


    def refill_show_room(self,module_gui:ModuleGUI,visible:bool=True,index:int=None,show_room_id:int=None):
        """
        To refill the show room after a module has been added to the pipeline.

        Attributes:
            module_gui (ModuleGUI): The module GUI that is being added to the pipeline.
            visible (bool): Whether the module is visible or not.
            index (int): To add it at the same position as the old module was placed in the show_room_modules list(import for change the pages).
            show_room_id (int): So the refilled module has the same id as the one added to the pipeline.
        """
        new_module_gui = ModuleGUI(self, module_gui.module_type, x=SPACING_X + SHOWROOM_PADDING_X / 2, y=module_gui.show_offset_y, show_mode=True, visible=visible, index=index,id_number=show_room_id)
        self.page_stack.controls.insert(3,new_module_gui)
        self.page_stack.update()
        self.update_all_port_icons()

    def build_show_room(self,page_stack:ft.Stack):
        """
        Builds the show room for the PipelineBuildingTool(ExpertMode).
        Its placed at the top left corner to add modules to the pipeline.
        """
        def _add_show_room_module(module_type: ModuleType, x: float, y: float, visible: bool = True, show_room_id: int = None):
            module_gui = ModuleGUI(self, module_type, x, y, True, visible, id_number=show_room_id)
            self.page_stack.controls.insert(2, module_gui)
            return module_gui

        self.page_stack = page_stack
        x = SPACING_X + SHOWROOM_PADDING_X / 2
        y = SPACING_Y
        self.show_room_container = ft.Container(top=y - SPACING_Y / 2, left=SPACING_X, width=MODULE_WIDTH + SHOWROOM_PADDING_X, height=(self.show_room_size * MODULE_HEIGHT) + (self.show_room_size * SPACING_Y), bgcolor=MENU_COLOR, border_radius=ft.border_radius.all(10),blur=10)
        self.page_stack.controls.insert(1,self.show_room_container)
        self.show_room_max_page_number = math.ceil(len(ModuleType) / SHOWROOM_MODULE_COUNT)
        for i,module_type in enumerate(ModuleType):
            visible = i < SHOWROOM_MODULE_COUNT
            y_module = y+ (MODULE_HEIGHT + SPACING_Y) * (i%SHOWROOM_MODULE_COUNT)
            _add_show_room_module(module_type,x,y_module,visible,-1)
            _add_show_room_module(module_type,x,y_module,visible,-2)

    def change_show_room_page(self,page_number:int):
        """
        Changes the page of the show_room to the given page number.
        """
        self.show_room_page_number = page_number
        for i,module in enumerate(self.show_room_modules):
            module_page = (i//2) // SHOWROOM_MODULE_COUNT
            if module_page == self.show_room_page_number:
                module.visible = True
            else:
                module.visible = False
            module.update()

    def update_show_room(self):
        """
        Updates the show_room position.
        """
        self.show_room_container.left = SPACING_X
        self.show_room_container.top = SPACING_Y - SPACING_Y / 2
        self.show_room_container.update()
        for module in self.show_room_modules:
            module.left = SPACING_X + SHOWROOM_PADDING_X / 2
            module.update()

    def add_module(self,module_type: ModuleType,x: float = None,y: float = None,module_id: str = None,module_dict:dict=None):
        """
        Adds a module to the PipelineGUI.
        """
        id_number = int(module_id.removeprefix(module_type.value.gui_config().name)) if module_id is not None else None
        module_gui = ModuleGUI(self,module_type,x,y,id_number=id_number,module_dict=module_dict)
        self.controls.append(module_gui)
        self.update()
        return module_gui

    def set_in_background(self, module_gui:ModuleGUI,behind_delete=False):
        """
        Move a Module from it current position in the stack to the deepest intended level.
        Attribute:
            behind_delete: if the module should be render behind the delete buttons.
        """
        if module_gui in self.controls:
            self.controls.remove(module_gui)
            if behind_delete:
                self.controls.insert(1, module_gui)
            else:
                self.controls.remove(self.delete_stack)
                self.controls.insert(1, self.delete_stack)
                self.controls.insert(2, module_gui)
            self.update()

    def remove_module(self,module_id: str):
        """
        Removes a module from the pipeline.
        """
        gui_module = self.modules.pop(module_id)
        self.controls.remove(gui_module)
        self.pipeline.remove_module(gui_module.module)
        del gui_module
        self.update()

    def toggle_all_module_detection(self,module_id: str):
        """
        Toggles whether all modules can only be moved and are ready to be clicked to connect connections or
        are in the 'normal' mode where any tool is available.
        """
        self.source_module = module_id
        self.transmitting_ports = []
        for module in self.modules.values():
            if module.module_id != module_id:
                module.toggle_detection()
                self.update()

    def enables_all_stuck_in_running(self):
        """
        Disables the waiting/execution mode for all modules when some event happens that the pipeline terminated early.
        """
        for module in self.modules.values():
            self.lines_gui.update_delete_buttons(module)
            module.enable_tools()
            module.delete_button.visible = True
            module.delete_button.update()
            module.executing_button.visible = False
            module.executing_button.update()
            module.waiting_button.visible = False
            module.waiting_button.update()
        self.check_for_valid_all_modules()

    def check_for_valid_all_modules(self):
        """
        Checks all modules if its valid to build a connection to the currently selected source module.
        """
        for target_module_gui in self.modules.values():
            self.check_for_valid(target_module_gui.module_id)

    def check_for_valid(self,module_id: str):
        """
        Checks for a module associated with the given module_id if its valid to build a connection to the currently selected source module.
        """
        target_module_gui = self.modules[module_id]
        if (target_module_gui.module_id not in self.pipeline.run_order and target_module_gui.module_id != self.pipeline.executing) or not self.pipeline.running:
            if target_module_gui.module_id != self.source_module:
                valid = True
                existing_pipe = self.pipeline.check_connections(
                    self.source_module, target_module_gui.module_id) if self.source_module != "" else None
                if existing_pipe is None:
                    valid = True
                elif any(port in existing_pipe.ports for port in self.transmitting_ports) or existing_pipe.source_module.module_id != self.source_module or existing_pipe.target_module.module_id != target_module_gui.module_id:
                    valid = False
                if all(k in target_module_gui.module.inputs for k in
                       self.transmitting_ports) and self.transmitting_ports != [] and valid and not (
                self.pipeline.check_ports_occupied(target_module_gui.module_id, self.transmitting_ports)):
                    target_module_gui.set_valid()
                else:
                    target_module_gui.set_invalid()

    def update_all_port_icons(self):
        """
        Updates for all modules the port icons.
        """
        for module_gui in self.modules.values():
            module_gui.update_port_icons()
